from itertools import permutations
from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.constants import (
    EMPTY_ADDRESS,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
    ChannelEvent,
    ChannelInfoIndex,
    ChannelState,
    ParticipantInfoIndex,
)
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    LOCKSROOT_OF_NO_LOCKS,
    NONEXISTENT_LOCKSROOT,
    NOT_ADDRESS,
    call_and_transact,
)
from raiden_contracts.utils.events import check_channel_opened

from .utils import get_participants_hash
from .utils.blockchain import mine_blocks


def test_open_channel_call(token_network: Contract, get_accounts: Callable) -> None:
    """Calling openChannel() with various wrong arguments"""
    (A, B) = get_accounts(2)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN + 10

    # Validation failure with a negative settle timeout
    with pytest.raises(ValidationError):
        token_network.functions.openChannel(A, B, -3)

    # Validation failure with the number zero instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.openChannel(0x0, B, settle_timeout)

    # Validation failure with the empty string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.openChannel("", B, settle_timeout)

    # Validation failure with an odd-length string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.openChannel(NOT_ADDRESS, B, settle_timeout)

    # Validation failure with the number zero instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.openChannel(A, 0x0, settle_timeout)

    # Validation failure with the empty string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.openChannel(A, "", settle_timeout)

    # Validation failure with an odd-length string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.openChannel(A, NOT_ADDRESS, settle_timeout)

    # Transaction failure with the zero address
    with pytest.raises(TransactionFailed, match="TN: participant address zero"):
        token_network.functions.openChannel(EMPTY_ADDRESS, B, settle_timeout).call()

    # Transaction failure with the zero address
    with pytest.raises(TransactionFailed, match="TN: partner address zero"):
        token_network.functions.openChannel(A, EMPTY_ADDRESS, settle_timeout).call()

    # Cannot open a channel between 2 participants with the same address
    with pytest.raises(TransactionFailed, match="TN: identical addresses"):
        token_network.functions.openChannel(A, A, settle_timeout).call()

    # Cannot open a channel for one-too-small settle timeout
    with pytest.raises(TransactionFailed, match="TN: settle timeout < min"):
        token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MIN - 1).call()

    # Cannot open a channel for one-too-big settle timeout
    with pytest.raises(TransactionFailed, match="TN: settle timeout > max"):
        token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MAX + 1).call()


def test_max_1_channel(
    token_network: Contract, get_accounts: Callable, create_channel: Callable
) -> None:
    """For two participants, at most one channel can be opened"""
    (A, B) = get_accounts(2)
    create_channel(A, B, TEST_SETTLE_TIMEOUT_MIN)

    with pytest.raises(TransactionFailed, match="TN/open: channel exists for participants"):
        token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MIN).call()
    with pytest.raises(TransactionFailed, match="TN/open: channel exists for participants"):
        token_network.functions.openChannel(B, A, TEST_SETTLE_TIMEOUT_MIN).call()


def test_participants_hash(token_network: Contract, get_accounts: Callable) -> None:
    """getParticipantsHash() behaves as get_participants_hash"""
    (A, B) = get_accounts(2)

    AB_hash = get_participants_hash(A, B)
    assert token_network.functions.getParticipantsHash(A, B).call() == AB_hash
    assert token_network.functions.getParticipantsHash(B, A).call() == AB_hash


def test_participants_hash_equal(token_network: Contract, get_accounts: Callable) -> None:
    """getParticipantsHash() behaves as get_participants_hash on equal addresses"""
    (A,) = get_accounts(1)

    with pytest.raises(ValueError):
        get_participants_hash(A, A)
    with pytest.raises(TransactionFailed, match="TN: identical addresses"):
        token_network.functions.getParticipantsHash(A, A).call()


def test_counter(
    token_network: Contract, get_accounts: Callable, create_channel: Callable
) -> None:
    """Open three channels and observe states"""
    (A, B, C, D) = get_accounts(4)

    AB_hash = token_network.functions.getParticipantsHash(A, B).call()
    BC_hash = token_network.functions.getParticipantsHash(B, C).call()
    CD_hash = token_network.functions.getParticipantsHash(C, D).call()

    assert token_network.functions.channel_counter().call() == 0
    assert token_network.functions.participants_hash_to_channel_identifier(AB_hash).call() == 0
    assert token_network.functions.participants_hash_to_channel_identifier(BC_hash).call() == 0
    assert token_network.functions.participants_hash_to_channel_identifier(CD_hash).call() == 0
    assert token_network.functions.getChannelIdentifier(A, B).call() == 0

    # Create channel between A and B, counter increases
    create_channel(A, B)
    assert token_network.functions.channel_counter().call() == 1
    assert token_network.functions.participants_hash_to_channel_identifier(AB_hash).call() == 1
    assert token_network.functions.getChannelIdentifier(A, B).call() == 1

    # We still do not have a channel between B and C
    assert token_network.functions.getChannelIdentifier(B, C).call() == 0

    # Create channel between B and C, counter increases
    create_channel(B, C)
    assert token_network.functions.channel_counter().call() == 2
    assert token_network.functions.participants_hash_to_channel_identifier(BC_hash).call() == 2
    assert token_network.functions.getChannelIdentifier(B, C).call() == 2

    # We still do not have a channel between C and D
    assert token_network.functions.getChannelIdentifier(C, D).call() == 0

    # Create channel between C and D, counter increases
    create_channel(C, D)
    assert token_network.functions.channel_counter().call() == 3
    assert token_network.functions.participants_hash_to_channel_identifier(CD_hash).call() == 3
    assert token_network.functions.getChannelIdentifier(C, D).call() == 3


def test_state_channel_identifier_invalid(
    token_network: Contract, get_accounts: Callable, create_channel: Callable
) -> None:
    """getChannelInfo() returns the empty channel state for ont-too-big channelID"""
    (A, B, C) = get_accounts(3)
    channel_id = 0

    pairs = permutations([A, B, C], 2)
    for pair in pairs:
        (settle_block_number, state) = token_network.functions.getChannelInfo(
            channel_id, *pair
        ).call()
        assert settle_block_number == 0
        assert state == ChannelState.NONEXISTENT

    for pair in pairs:
        create_channel(*pair)
        (settle_block_number, state) = token_network.functions.getChannelInfo(0, *pair).call()
        assert settle_block_number > 0
        assert state == ChannelState.OPENED

    current_counter = token_network.functions.channel_counter().call()

    for pair in pairs:
        (settle_block_number, state) = token_network.functions.getChannelInfo(
            current_counter + 1, *pair
        ).call()
        assert settle_block_number == 0
        assert state == ChannelState.NONEXISTENT


def test_open_channel_state(token_network: Contract, get_accounts: Callable) -> None:
    """Observe the state of the channel after a openChannel() call"""
    (A, B) = get_accounts(2)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN + 10

    channel_counter = token_network.functions.channel_counter().call()
    participants_hash = token_network.functions.getParticipantsHash(A, B).call()

    assert (
        token_network.functions.participants_hash_to_channel_identifier(participants_hash).call()
        == 0
    )
    assert token_network.functions.getChannelIdentifier(A, B).call() == 0

    call_and_transact(token_network.functions.openChannel(A, B, settle_timeout))
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()

    assert token_network.functions.channel_counter().call() == channel_counter + 1
    assert (
        token_network.functions.participants_hash_to_channel_identifier(participants_hash).call()
        == channel_counter + 1
    )

    channel_info_response = token_network.functions.getChannelInfo(channel_identifier, A, B).call()
    settle_block_number = channel_info_response[ChannelInfoIndex.SETTLE_BLOCK]
    state = channel_info_response[ChannelInfoIndex.STATE]
    assert settle_block_number == settle_timeout
    assert state == ChannelState.OPENED

    response = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    A_deposit = response[ParticipantInfoIndex.DEPOSIT]
    A_withdrawn = response[ParticipantInfoIndex.WITHDRAWN]
    A_is_the_closer = response[ParticipantInfoIndex.IS_CLOSER]
    A_balance_hash = response[ParticipantInfoIndex.BALANCE_HASH]
    A_nonce = response[ParticipantInfoIndex.NONCE]
    A_locksroot = response[ParticipantInfoIndex.LOCKSROOT]
    A_locked_amount = response[ParticipantInfoIndex.LOCKED_AMOUNT]
    assert A_deposit == 0
    assert A_withdrawn == 0
    assert A_is_the_closer is False
    assert A_balance_hash == EMPTY_BALANCE_HASH
    assert A_nonce == 0
    assert A_locksroot == NONEXISTENT_LOCKSROOT
    assert A_locked_amount == 0

    (
        B_deposit,
        B_withdrawn,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        B_locksroot,
        B_locked_amount,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert B_deposit == 0
    assert B_withdrawn == 0
    assert B_is_the_closer is False
    assert B_balance_hash == EMPTY_BALANCE_HASH
    assert B_nonce == 0
    assert B_locksroot == NONEXISTENT_LOCKSROOT
    assert B_locked_amount == 0


def test_reopen_channel(
    web3: Web3,
    token_network: Contract,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Open a second channel after settling one"""
    (A, B) = get_accounts(2)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN

    call_and_transact(token_network.functions.openChannel(A, B, settle_timeout))
    channel_identifier1 = token_network.functions.getChannelIdentifier(A, B).call()
    channel_counter1 = token_network.functions.participants_hash_to_channel_identifier(
        get_participants_hash(A, B)
    ).call()

    # Opening twice fails
    with pytest.raises(TransactionFailed, match="TN/open: channel exists for participants"):
        token_network.functions.openChannel(A, B, settle_timeout).call()

    # Close channel
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier1)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier1,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    # Reopen Channel before settlement fails
    with pytest.raises(TransactionFailed, match="TN/open: channel exists for participants"):
        token_network.functions.openChannel(A, B, settle_timeout).call()

    # Settlement window must be over before settling the channel
    mine_blocks(web3, settle_timeout + 1)

    # Settle channel
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier1,
            A,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
            B,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )

    # Reopening the channel should work iff channel is settled
    call_and_transact(token_network.functions.openChannel(A, B, settle_timeout))
    channel_identifier2 = token_network.functions.getChannelIdentifier(A, B).call()
    assert channel_identifier2 != channel_identifier1
    assert (
        token_network.functions.participants_hash_to_channel_identifier(
            get_participants_hash(A, B)
        ).call()
        == channel_counter1 + 1
    )

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier2, A, B
    ).call()
    assert settle_block_number == settle_timeout
    assert state == ChannelState.OPENED

    (
        A_deposit,
        A_withdrawn,
        A_is_the_closer,
        A_balance_hash,
        A_nonce,
        A_locksroot,
        A_locked_amount,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier2, A, B).call()
    assert A_deposit == 0
    assert A_withdrawn == 0
    assert A_is_the_closer is False
    assert A_balance_hash == EMPTY_BALANCE_HASH
    assert A_nonce == 0
    assert A_locksroot == NONEXISTENT_LOCKSROOT
    assert A_locked_amount == 0

    (
        B_deposit,
        B_withdrawn,
        B_is_the_closer,
        B_balance_hash,
        B_nonce,
        B_locksroot,
        B_locked_amount,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier2, B, A).call()
    assert B_deposit == 0
    assert B_withdrawn == 0
    assert B_is_the_closer is False
    assert B_balance_hash == EMPTY_BALANCE_HASH
    assert B_nonce == 0
    assert B_locksroot == NONEXISTENT_LOCKSROOT
    assert B_locked_amount == 0


def test_open_channel_event(
    get_accounts: Callable, token_network: Contract, event_handler: Callable
) -> None:
    """A successful openChannel() causes an OPENED event"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)

    txn_hash = call_and_transact(
        token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MIN)
    )
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()

    ev_handler.add(
        txn_hash,
        ChannelEvent.OPENED,
        check_channel_opened(channel_identifier, A, B, TEST_SETTLE_TIMEOUT_MIN),
    )
    ev_handler.check()



from unittest.mock import Mock

import pytest
from hexbytes import HexBytes
from web3.types import TxData, TxReceipt, Wei

from raiden_contracts.utils.transaction import check_successful_tx


def test_check_successful_tx_with_status_zero() -> None:
    web3_mock = Mock()
    web3_mock.eth.get_transaction_receipt.return_value = {"blockNumber": 300, "status": 0}
    txid = HexBytes("abcdef")
    with pytest.raises(ValueError):
        check_successful_tx(web3=web3_mock, txid=txid)
    web3_mock.eth.get_transaction_receipt.assert_called_with(txid)
    web3_mock.eth.get_transaction.assert_called_with(txid)


def test_check_successful_tx_with_nonexistent_status() -> None:
    """check_successful_tx() with a receipt without status field should raise a KeyError"""
    web3_mock = Mock()
    web3_mock.eth.get_transaction_receipt.return_value = {"blockNumber": 300}
    txid = HexBytes("abcdef")
    with pytest.raises(KeyError):
        check_successful_tx(web3=web3_mock, txid=txid)
    web3_mock.eth.get_transaction_receipt.assert_called_with(txid)
    web3_mock.eth.get_transaction.assert_called_with(txid)


def test_check_successful_tx_with_gas_completely_used() -> None:
    web3_mock = Mock()
    gas = 30000
    web3_mock.eth.get_transaction_receipt.return_value = {
        "blockNumber": 300,
        "status": 1,
        "gasUsed": gas,
    }
    web3_mock.eth.get_transaction.return_value = {"gas": gas}
    txid = HexBytes("abcdef")
    with pytest.raises(ValueError):
        check_successful_tx(web3=web3_mock, txid=txid)
    web3_mock.eth.get_transaction_receipt.assert_called_with(txid)
    web3_mock.eth.get_transaction.assert_called_with(txid)


def test_check_successful_tx_successful_case() -> None:
    web3_mock = Mock()
    gas = 30000
    receipt = TxReceipt({"blockNumber": 300, "status": 1, "gasUsed": gas - 10})  # type: ignore
    web3_mock.eth.get_transaction_receipt.return_value = receipt
    txinfo = TxData({"gas": Wei(gas)})
    web3_mock.eth.get_transaction.return_value = txinfo
    txid = HexBytes("abcdef")
    assert check_successful_tx(web3=web3_mock, txid=txid) == (receipt, txinfo)
    web3_mock.eth.get_transaction_receipt.assert_called_with(txid)
    web3_mock.eth.get_transaction.assert_called_with(txid)



from collections import namedtuple
from enum import Enum, IntEnum
from typing import Dict

from eth_typing import ChecksumAddress, HexAddress, HexStr
from eth_utils import keccak

from raiden_contracts.utils.type_aliases import ChainID, Locksroot

# The last digit is supposed to be zero always. See `RELEASE.rst`.
CONTRACTS_VERSION = "0.39.0"
BESPIN_VERSION = "0.37.0"
ALDERAAN_VERSION = "0.37.0"

PRECOMPILED_DATA_FIELDS = ["abi", "bin", "bin-runtime", "metadata"]

# Contract names
CONTRACT_HUMAN_STANDARD_TOKEN = "HumanStandardToken"
CONTRACT_TOKEN_NETWORK_REGISTRY = "TokenNetworkRegistry"
CONTRACT_TOKEN_NETWORK = "TokenNetwork"
CONTRACT_SECRET_REGISTRY = "SecretRegistry"
CONTRACT_CUSTOM_TOKEN = "CustomToken"
CONTRACT_CUSTOM_TOKEN_NO_DECIMALS = "CustomTokenNoDecimals"
CONTRACT_MONITORING_SERVICE = "MonitoringService"
CONTRACT_RAIDEN_SERVICE_BUNDLE = "ServiceRegistry"  # old name for compatibility, deprecated
CONTRACT_SERVICE_REGISTRY = "ServiceRegistry"
CONTRACT_USER_DEPOSIT = "UserDeposit"
CONTRACT_ONE_TO_N = "OneToN"
CONTRACT_DEPOSIT = "Deposit"

# Timeouts
TEST_SETTLE_TIMEOUT_MIN = 5
TEST_SETTLE_TIMEOUT_MAX = 100000

DEPLOY_SETTLE_TIMEOUT_MIN = 500  # ~ 2 hours
DEPLOY_SETTLE_TIMEOUT_MAX = 555428  # ~ 3 months

# Temporary deposit limits for the Red Eyes release in WEI
MAX_ETH_CHANNEL_PARTICIPANT = int(0.075 * 10 ** 18)
MAX_ETH_TOKEN_NETWORK = int(250 * 10 ** 18)

# Special hashes
LOCKSROOT_OF_NO_LOCKS = Locksroot(keccak(b""))
EMPTY_ADDRESS = ChecksumAddress(HexAddress(HexStr("0x0000000000000000000000000000000000000000")))

# Event names
# TokenNetworkRegistry
EVENT_TOKEN_NETWORK_CREATED = "TokenNetworkCreated"

# TokenNetwork
EVENT_DEPRECATION_SWITCH = "DeprecationSwitch"

# SecretRegistry
EVENT_SECRET_REVEALED = "SecretRevealed"

# EndpointRegistry
EVENT_ADDRESS_REGISTERED = "AddressRegistered"

# ServiceRegistry
EVENT_REGISTERED_SERVICE = "RegisteredService"


class ChannelEvent(str, Enum):
    OPENED = "ChannelOpened"
    DEPOSIT = "ChannelNewDeposit"
    WITHDRAW = "ChannelWithdraw"
    BALANCE_PROOF_UPDATED = "NonClosingBalanceProofUpdated"
    CLOSED = "ChannelClosed"
    SETTLED = "ChannelSettled"
    UNLOCKED = "ChannelUnlocked"
    DEPRECATED = "DeprecationSwitch"


class ChannelInfoIndex(IntEnum):
    """Index for accessing fields in return information from TokenNetwork.getChannelInfo."""

    SETTLE_BLOCK = 0
    STATE = 1


class ParticipantInfoIndex(IntEnum):
    """Index for accessing fields in return value of TokenNetwork.getChannelParticipantInfo."""

    DEPOSIT = 0
    WITHDRAWN = 1
    IS_CLOSER = 2
    BALANCE_HASH = 3
    NONCE = 4
    LOCKSROOT = 5
    LOCKED_AMOUNT = 6


class ChannelState(IntEnum):
    """Meaning of values returned by TokenNetwork.getChannelInfo[ChannelInfoIndex.STATE]"""

    NONEXISTENT = 0
    OPENED = 1
    CLOSED = 2
    SETTLED = 3
    REMOVED = 4


class MessageTypeId(IntEnum):
    """Message types, as used by the TokenNetwork contract"""

    BALANCE_PROOF = 1
    BALANCE_PROOF_UPDATE = 2
    WITHDRAW = 3
    COOPERATIVE_SETTLE = 4
    IOU = 5
    MSReward = 6


class MonitoringServiceEvent(str, Enum):
    """Message types used by MonitoringService contract"""

    NEW_BALANCE_PROOF_RECEIVED = "NewBalanceProofReceived"
    REWARD_CLAIMED = "RewardClaimed"


class UserDepositEvent(str, Enum):
    """Message types used by UserDeposit contract"""

    BALANCE_REDUCED = "BalanceReduced"
    WITHDRAW_PLANNED = "WithdrawPlanned"


class OneToNEvent(str, Enum):
    """Message types used by OneToN contract"""

    CLAIMED = "Claimed"


class DeploymentModule(Enum):
    """Groups of contracts that are deployed together"""

    RAIDEN = "raiden"
    SERVICES = "services"
    ALL = "all"


# Network configurations
START_QUERY_BLOCK_KEY = "DefaultStartBlock"

ID_TO_CHAINNAME: Dict[ChainID, str] = {
    ChainID(-5): "goerli_unstable",
    ChainID(1): "mainnet",
    ChainID(3): "ropsten",
    ChainID(4): "rinkeby",
    ChainID(5): "goerli",
    ChainID(42): "kovan",
    ChainID(627): "smoketest",
}

CHAINNAME_TO_ID: Dict[str, ChainID] = {name: id for id, name in ID_TO_CHAINNAME.items()}

# ContractNames


ContractListEntry = namedtuple("ContractListEntry", "module name")

CONTRACT_LIST = [
    ContractListEntry(module=DeploymentModule.RAIDEN, name=CONTRACT_SECRET_REGISTRY),
    ContractListEntry(module=DeploymentModule.RAIDEN, name=CONTRACT_TOKEN_NETWORK_REGISTRY),
    ContractListEntry(module=DeploymentModule.RAIDEN, name=CONTRACT_TOKEN_NETWORK),
    ContractListEntry(module=DeploymentModule.SERVICES, name=CONTRACT_SERVICE_REGISTRY),
    ContractListEntry(module=DeploymentModule.SERVICES, name=CONTRACT_MONITORING_SERVICE),
    ContractListEntry(module=DeploymentModule.SERVICES, name=CONTRACT_ONE_TO_N),
    ContractListEntry(module=DeploymentModule.SERVICES, name=CONTRACT_USER_DEPOSIT),
]



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from eth_typing.evm import HexAddress
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    MAX_ETH_CHANNEL_PARTICIPANT,
    MAX_ETH_TOKEN_NETWORK,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
    ParticipantInfoIndex,
)
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS, UINT256_MAX


def test_register_three_but_not_four(
    web3: Web3,
    get_token_network_registry: Callable,
    secret_registry_contract: Contract,
    custom_token_factory: Callable,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> None:
    """Check that TokenNetworkRegistry observes the max number of tokens"""
    token_network_registry = get_token_network_registry(
        _secret_registry_address=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
        _max_token_networks=3,
    )
    assert token_network_registry.functions.max_token_networks().call() == 3
    token0 = custom_token_factory()
    token1 = custom_token_factory()
    token2 = custom_token_factory()
    token3 = custom_token_factory()
    token4 = custom_token_factory()
    call_and_transact(
        token_network_registry.functions.createERC20TokenNetwork(
            token0.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        )
    )
    # Registering the same token again must fail
    with pytest.raises(TransactionFailed, match="TNR: token already registered"):
        call_and_transact(
            token_network_registry.functions.createERC20TokenNetwork(
                token0.address,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            )
        )
    call_and_transact(
        token_network_registry.functions.createERC20TokenNetwork(
            token1.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        )
    )
    call_and_transact(
        token_network_registry.functions.createERC20TokenNetwork(
            token2.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        )
    )
    with pytest.raises(TransactionFailed, match="TNR: registry full"):
        token_network_registry.functions.createERC20TokenNetwork(
            token3.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ).call()

    # Now remove the limit and try again
    call_and_transact(token_network_registry.functions.removeLimits(), {"from": DEPLOYER_ADDRESS})
    with pytest.raises(TransactionFailed, match="TNR: limits must be set to MAX_INT"):
        call_and_transact(
            token_network_registry.functions.createERC20TokenNetwork(
                token4.address,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            )
        )
    call_and_transact(
        token_network_registry.functions.createERC20TokenNetwork(
            token4.address,
            UINT256_MAX,
            UINT256_MAX,
        )
    )


def test_channel_participant_deposit_limit_value(token_network: Contract) -> None:
    """Check the channel participant deposit limit"""
    limit = token_network.functions.channel_participant_deposit_limit().call()
    assert limit == MAX_ETH_CHANNEL_PARTICIPANT


def test_network_deposit_limit_value(token_network: Contract) -> None:
    """Check the token network deposit limit"""
    limit = token_network.functions.token_network_deposit_limit().call()
    assert limit == MAX_ETH_TOKEN_NETWORK


def test_participant_deposit_limit(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    assign_tokens: Callable,
) -> None:
    """Observe failure to deposit a bit more tokens than the participant deposit limit"""
    (A, B) = get_accounts(2)
    deposit_A = 100000
    deposit_B = 100000
    channel_identifier = create_channel(A, B)[0]
    assign_tokens(A, MAX_ETH_CHANNEL_PARTICIPANT + 10)
    assign_tokens(B, MAX_ETH_CHANNEL_PARTICIPANT + 10)

    with pytest.raises(TransactionFailed, match="TN/deposit: deposit limit reached"):
        token_network.functions.setTotalDeposit(
            channel_identifier, A, MAX_ETH_CHANNEL_PARTICIPANT + 1, B
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/deposit: deposit limit reached"):
        token_network.functions.setTotalDeposit(
            channel_identifier, B, MAX_ETH_CHANNEL_PARTICIPANT + 1, A
        ).call({"from": B})

    # Deposit some tokens, under the limit
    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, A, deposit_A, B),
        {"from": A},
    )
    info_A = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    assert info_A[ParticipantInfoIndex.DEPOSIT] == deposit_A

    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, B, deposit_B, A),
        {"from": B},
    )
    info_B = token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()
    assert info_B[ParticipantInfoIndex.DEPOSIT] == deposit_B

    with pytest.raises(TransactionFailed, match="TN/deposit: deposit limit reached"):
        token_network.functions.setTotalDeposit(
            channel_identifier, A, MAX_ETH_CHANNEL_PARTICIPANT + 1, B
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/deposit: deposit limit reached"):
        token_network.functions.setTotalDeposit(
            channel_identifier, B, MAX_ETH_CHANNEL_PARTICIPANT + 1, A
        ).call({"from": B})

    call_and_transact(
        token_network.functions.setTotalDeposit(
            channel_identifier, A, MAX_ETH_CHANNEL_PARTICIPANT, B
        ),
        {"from": A},
    )
    call_and_transact(
        token_network.functions.setTotalDeposit(
            channel_identifier, B, MAX_ETH_CHANNEL_PARTICIPANT, A
        ),
        {"from": B},
    )

    # Now disable the limits and try again
    with pytest.raises(TransactionFailed, match="Can only be called by"):
        call_and_transact(token_network.functions.removeLimits())
    call_and_transact(token_network.functions.removeLimits(), {"from": DEPLOYER_ADDRESS})
    token_network.functions.setTotalDeposit(
        channel_identifier, B, MAX_ETH_CHANNEL_PARTICIPANT + 1, A
    ).call({"from": B})


@pytest.mark.skip(reason="Only for local testing, otherwise it takes too much time to run.")
def test_network_deposit_limit(
    create_account: Callable,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    assign_tokens: Callable,
) -> None:
    last_deposit = 100

    # ! Only for testing, otherwise we need 1300+ channels and test needs a lot of time to complete
    # ! The token_network_deposit_limit also needs to be changed for this to work
    MAX_ETH_TOKEN_NETWORK_TESTING = int(1 * 10 ** 18)

    def remaining() -> int:
        return (
            MAX_ETH_TOKEN_NETWORK_TESTING
            - custom_token.functions.balanceOf(token_network.address).call()
            - last_deposit
        )

    def send_remaining(
        channel_identifier: int, participant1: HexAddress, participant2: HexAddress
    ) -> None:
        remaining_to_reach_limit = remaining()
        assign_tokens(participant1, remaining_to_reach_limit)
        call_and_transact(
            token_network.functions.setTotalDeposit(
                channel_identifier, participant1, remaining_to_reach_limit, participant2
            ),
            {"from": participant1},
        )

    remaining_to_reach_limit = remaining()
    while remaining_to_reach_limit > 0:
        A = create_account()
        B = create_account()
        assign_tokens(A, MAX_ETH_CHANNEL_PARTICIPANT)
        assign_tokens(B, MAX_ETH_CHANNEL_PARTICIPANT)
        channel_identifier = create_channel(A, B)[0]

        try:
            call_and_transact(
                token_network.functions.setTotalDeposit(
                    channel_identifier, A, MAX_ETH_CHANNEL_PARTICIPANT, B
                ),
                {"from": A},
            )
        except TransactionFailed:
            send_remaining(channel_identifier, A, B)
            break

        try:
            call_and_transact(
                token_network.functions.setTotalDeposit(
                    channel_identifier, B, MAX_ETH_CHANNEL_PARTICIPANT, A
                ),
                {"from": B},
            )
        except TransactionFailed:
            send_remaining(channel_identifier, B, A)
            break

        remaining_to_reach_limit = remaining()

    assert (
        MAX_ETH_TOKEN_NETWORK_TESTING
        - custom_token.functions.balanceOf(token_network.address).call()
        == last_deposit
    )

    A = create_account()
    B = create_account()
    assign_tokens(A, last_deposit + 10)
    channel_identifier = create_channel(A, B)[0]

    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, A, last_deposit, B),
        {"from": A},
    )

    # After token network limit is reached, we cannot deposit anymore tokens in existent channels
    with pytest.raises(TransactionFailed, match="abc"):
        token_network.functions.setTotalDeposit(channel_identifier, A, 1, B).call({"from": A})

    # After token network limit is reached, we cannot open new channels
    C = create_account()
    D = create_account()
    with pytest.raises(TransactionFailed, match="abc"):
        create_channel(C, D)

    # Now disable the limits and try again
    call_and_transact(token_network.functions.removeLimits(), {"from": DEPLOYER_ADDRESS})
    token_network.functions.setTotalDeposit(channel_identifier, A, 1, B).call({"from": A})



from typing import Union

from coincurve import PrivateKey, PublicKey
from eth_typing import ChecksumAddress, HexStr
from eth_utils import keccak, to_bytes, to_checksum_address

from raiden_contracts.utils.type_aliases import PrivateKey as ContractsPrivateKey


def sign(privkey: ContractsPrivateKey, msg_hash: bytes, v: int = 0) -> bytes:
    if not isinstance(msg_hash, bytes):
        raise TypeError("sign(): msg_hash is not an instance of bytes")
    if len(msg_hash) != 32:
        raise ValueError("sign(): msg_hash has to be exactly 32 bytes")
    if not isinstance(privkey, bytes):
        raise TypeError("sign(): privkey is not an instance of bytes")
    if v not in {0, 27}:
        raise ValueError(f"sign(): got v = {v} expected 0 or 27.")

    pk = PrivateKey(privkey)
    sig: bytes = pk.sign_recoverable(msg_hash, hasher=None)
    assert len(sig) == 65

    pub = pk.public_key
    recovered = PublicKey.from_signature_and_message(sig, msg_hash, hasher=None)
    assert pub == recovered

    sig = sig[:-1] + bytes([sig[-1] + v])

    return sig


def private_key_to_address(
    private_key: Union[PrivateKey, ContractsPrivateKey, bytes, str]
) -> ChecksumAddress:
    """Converts a private key to an Ethereum address."""
    if isinstance(private_key, str):
        pk = PrivateKey(to_bytes(hexstr=HexStr(private_key)))
    elif isinstance(private_key, bytes):
        pk = PrivateKey(private_key)
    else:
        pk = private_key

    return public_key_to_address(pk.public_key)


def public_key_to_address(public_key: Union[PublicKey, bytes]) -> ChecksumAddress:
    """Converts a public key to an Ethereum address."""
    if isinstance(public_key, PublicKey):
        public_key = public_key.format(compressed=False)
    assert isinstance(public_key, bytes)
    return to_checksum_address(keccak(public_key[1:])[-20:])



from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Generator

import pytest

from raiden_contracts.contract_source_manager import ContractSourceManager, contracts_source_path


@pytest.fixture(scope="session")
def contract_source_manager() -> ContractSourceManager:
    return ContractSourceManager(contracts_source_path(contracts_version=None))


@pytest.fixture(scope="session")
def contracts_manager(contract_source_manager: ContractSourceManager) -> Generator:
    with NamedTemporaryFile() as target_path:
        yield contract_source_manager.compile_contracts(Path(target_path.name))



from typing import Callable, Dict, List

import pytest
from eth_abi import encode_single
from eth_tester.exceptions import TransactionFailed
from eth_typing import HexAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    CONTRACT_MONITORING_SERVICE,
    LOCKSROOT_OF_NO_LOCKS,
    MessageTypeId,
    MonitoringServiceEvent,
)
from raiden_contracts.tests.utils import (
    DEPLOYER_ADDRESS,
    EMPTY_HEXADDRESS,
    SERVICE_DEPOSIT,
    call_and_transact,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.proofs import sign_reward_proof
from raiden_contracts.utils.type_aliases import TokenAmount

REWARD_AMOUNT = TokenAmount(10)


@pytest.fixture
def ms_address(
    get_accounts: Callable, custom_token: Contract, service_registry: Contract
) -> HexAddress:
    (ms,) = get_accounts(1)

    # register MS in the ServiceRegistry contract
    call_and_transact(custom_token.functions.mint(2 * SERVICE_DEPOSIT), {"from": ms})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT),
        {"from": ms},
    )
    call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": ms})

    return ms


@pytest.fixture
def setup_monitor_data(
    get_accounts: Callable,
    deposit_to_udc: Callable,
    create_channel: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    token_network: Contract,
    ms_address: HexAddress,
    get_private_key: Callable,
) -> Callable:
    def f(monitoring_service_contract: Contract) -> Dict:
        # Create two parties and a channel between them
        (A, B) = get_accounts(2, privkeys=[bytes([1] * 32), bytes([2] * 32)])
        deposit_to_udc(B, REWARD_AMOUNT)
        channel_identifier = create_channel(A, B)[0]

        # Create balance proofs
        balance_proof_A = create_balance_proof(
            channel_identifier, B, transferred_amount=10, nonce=1
        )
        balance_proof_B = create_balance_proof(
            channel_identifier, A, transferred_amount=20, nonce=2
        )

        # Add signatures by non_closing_participant
        closing_signature_A = create_balance_proof_countersignature(
            participant=A,
            channel_identifier=channel_identifier,
            msg_type=MessageTypeId.BALANCE_PROOF,
            **balance_proof_A._asdict(),
        )
        non_closing_signature_B = create_balance_proof_countersignature(
            participant=B,
            channel_identifier=channel_identifier,
            msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
            **balance_proof_B._asdict(),
        )
        reward_proof_signature = sign_reward_proof(
            privatekey=get_private_key(B),
            monitoring_service_contract_address=monitoring_service_contract.address,
            chain_id=token_network.functions.chain_id().call(),
            token_network_address=token_network.address,
            non_closing_participant=B,
            non_closing_signature=non_closing_signature_B,
            reward_amount=REWARD_AMOUNT,
        )

        # close channel
        call_and_transact(
            token_network.functions.closeChannel(
                channel_identifier,
                B,
                A,
                *balance_proof_A._asdict().values(),
                closing_signature_A,
            ),
            {"from": A},
        )

        # calculate when this MS is allowed to monitor
        first_allowed = monitoring_service_contract.functions.firstBlockAllowedToMonitorChannel(
            token_network=token_network.address,
            channel_identifier=channel_identifier,
            closing_participant=A,
            non_closing_participant=B,
            monitoring_service_address=ms_address,
        ).call()

        # return args for `monitor` function
        return {
            "participants": (A, B),
            "balance_proof_A": balance_proof_A,
            "balance_proof_B": balance_proof_B,
            "non_closing_signature": non_closing_signature_B,
            "reward_proof_signature": reward_proof_signature,
            "channel_identifier": channel_identifier,
            "first_allowed": first_allowed,
        }

    return f


@pytest.fixture
def monitor_data(setup_monitor_data: Callable, monitoring_service_external: Contract) -> Dict:
    return setup_monitor_data(monitoring_service_external)


@pytest.fixture
def monitor_data_internal(
    setup_monitor_data: Callable, monitoring_service_internals: Contract
) -> Dict:
    return setup_monitor_data(monitoring_service_internals)


@pytest.mark.parametrize("with_settle", [True, False])
def test_claimReward_with_settle_call(
    token_network: Contract,
    monitoring_service_external: Contract,
    user_deposit_contract: Contract,
    event_handler: Callable,
    monitor_data: Dict,
    ms_address: HexAddress,
    web3: Web3,
    with_settle: bool,
) -> None:
    A, B = monitor_data["participants"]
    channel_identifier = monitor_data["channel_identifier"]

    # wait until MS is allowed to monitor
    mine_blocks(web3, monitor_data["first_allowed"] - web3.eth.block_number)

    # MS updates closed channel on behalf of B
    txn_hash = call_and_transact(
        monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"]._asdict().values(),
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ),
        {"from": ms_address},
    )

    # claiming before settlement timeout must fail
    with pytest.raises(TransactionFailed, match="channel not settled yet"):
        monitoring_service_external.functions.claimReward(
            channel_identifier, token_network.address, A, B
        ).call({"from": ms_address})

    # Settle channel after settle_timeout elapsed
    mine_blocks(web3, 4)
    if with_settle:
        call_and_transact(
            token_network.functions.settleChannel(
                channel_identifier,
                B,  # participant_B
                10,  # participant_B_transferred_amount
                0,  # participant_B_locked_amount
                LOCKSROOT_OF_NO_LOCKS,  # participant_B_locksroot
                A,  # participant_A
                20,  # participant_A_transferred_amount
                0,  # participant_A_locked_amount
                LOCKSROOT_OF_NO_LOCKS,  # participant_A_locksroot
            )
        )

    # Claim reward for MS
    call_and_transact(
        monitoring_service_external.functions.claimReward(
            channel_identifier, token_network.address, A, B
        ),
        {"from": ms_address},
    )

    # Check REWARD_CLAIMED event
    reward_identifier = Web3.keccak(
        encode_single("uint256", channel_identifier) + Web3.toBytes(hexstr=token_network.address)
    )
    ms_ev_handler = event_handler(monitoring_service_external)
    ms_ev_handler.assert_event(
        txn_hash,
        MonitoringServiceEvent.REWARD_CLAIMED,
        dict(
            ms_address=ms_address,
            amount=REWARD_AMOUNT,
            reward_identifier=reward_identifier,
        ),
    )

    # Check that MS balance has increased by claiming the reward
    ms_balance_after_reward = user_deposit_contract.functions.balances(ms_address).call()
    assert ms_balance_after_reward == REWARD_AMOUNT


def test_monitor(
    token_network: Contract,
    monitoring_service_external: Contract,
    monitor_data: Dict,
    ms_address: HexAddress,
    event_handler: Callable,
    web3: Web3,
) -> None:
    A, B = monitor_data["participants"]

    # UpdateNonClosingBalanceProof is tested separately, so we assume that all
    # parameters passed to it are handled correctly.

    # Changing reward amount must lead to a failure during reward signature check.
    with pytest.raises(TransactionFailed, match="Reward proof with wrong non_closing_participant"):
        call_and_transact(
            monitoring_service_external.functions.monitor(
                A,
                B,
                *monitor_data["balance_proof_B"]._asdict().values(),
                monitor_data["non_closing_signature"],
                REWARD_AMOUNT + 1,
                token_network.address,
                monitor_data["reward_proof_signature"],
            ),
            {"from": ms_address},
        )

    # monitoring too early must fail
    with pytest.raises(TransactionFailed, match="not allowed to monitor"):
        assert web3.eth.block_number < monitor_data["first_allowed"]
        call_and_transact(
            monitoring_service_external.functions.monitor(
                A,
                B,
                *monitor_data["balance_proof_B"]._asdict().values(),
                monitor_data["non_closing_signature"],
                REWARD_AMOUNT,
                token_network.address,
                monitor_data["reward_proof_signature"],
            ),
            {"from": ms_address},
        )

    # wait until MS is allowed to monitor
    mine_blocks(web3, monitor_data["first_allowed"] - web3.eth.block_number)

    # successful monitor call
    txn_hash = call_and_transact(
        monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"]._asdict().values(),
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ),
        {"from": ms_address},
    )

    # NEW_BALANCE_PROOF_RECEIVED must get emitted
    ms_ev_handler = event_handler(monitoring_service_external)
    ms_ev_handler.assert_event(
        txn_hash,
        MonitoringServiceEvent.NEW_BALANCE_PROOF_RECEIVED,
        dict(
            token_network_address=token_network.address,
            channel_identifier=monitor_data["channel_identifier"],
            reward_amount=REWARD_AMOUNT,
            nonce=monitor_data["balance_proof_B"].nonce,
            ms_address=ms_address,
            raiden_node_address=B,
        ),
    )


def test_monitor_by_unregistered_service(
    token_network: Contract,
    monitoring_service_external: Contract,
    monitor_data: Dict,
    ms_address: HexAddress,
    web3: Web3,
) -> None:
    A, B = monitor_data["participants"]

    # wait until MS is allowed to monitor
    mine_blocks(web3, monitor_data["first_allowed"] - web3.eth.block_number)

    # only registered service providers may call `monitor`
    with pytest.raises(TransactionFailed, match="service not registered"):
        monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"]._asdict().values(),
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ).call({"from": B})

    # See a success to make sure the above failure is not spurious
    call_and_transact(
        monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"]._asdict().values(),
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ),
        {"from": ms_address},
    )


def test_monitor_on_wrong_token_network_registry(
    token_network_in_another_token_network_registry: Contract,
    monitoring_service_external: Contract,
    monitor_data: Dict,
    ms_address: HexAddress,
    web3: Web3,
) -> None:
    A, B = monitor_data["participants"]

    # wait until MS is allowed to monitor
    mine_blocks(web3, monitor_data["first_allowed"] - web3.eth.block_number)

    # monitor() call fails because the TokenNetwork is not registered on the
    # supposed TokenNetworkRegistry
    with pytest.raises(TransactionFailed, match="Unknown TokenNetwork"):
        call_and_transact(
            monitoring_service_external.functions.monitor(
                A,
                B,
                *monitor_data["balance_proof_B"]._asdict().values(),
                monitor_data["non_closing_signature"],
                REWARD_AMOUNT,
                token_network_in_another_token_network_registry.address,
                monitor_data["reward_proof_signature"],
            ),
            {"from": ms_address},
        )


def test_updateReward(
    monitoring_service_internals: Contract,
    ms_address: HexAddress,
    token_network: Contract,
    monitor_data_internal: Dict,
) -> None:
    A, B = monitor_data_internal["participants"]
    reward_identifier = Web3.keccak(
        encode_single("uint256", monitor_data_internal["channel_identifier"])
        + Web3.toBytes(hexstr=token_network.address)
    )

    def update_with_nonce(nonce: int) -> None:
        call_and_transact(
            monitoring_service_internals.functions.updateRewardPublic(
                token_network.address,
                A,
                B,
                REWARD_AMOUNT,
                nonce,
                ms_address,
                monitor_data_internal["non_closing_signature"],
                monitor_data_internal["reward_proof_signature"],
            ),
            {"from": ms_address},
        )

    # normal first call succeeds
    update_with_nonce(2)
    assert monitoring_service_internals.functions.rewardNonce(reward_identifier).call() == 2

    # calling again with same nonce fails
    with pytest.raises(TransactionFailed, match="stale nonce"):
        update_with_nonce(2)

    # calling again with higher nonce succeeds
    update_with_nonce(3)
    assert monitoring_service_internals.functions.rewardNonce(reward_identifier).call() == 3


def test_pureFirstAllowedBlock(monitoring_service_external: Contract) -> None:
    def call(addresses: List[int], closed_at_block: int = 1000, settle_timeout: int = 100) -> int:
        first_allowed = monitoring_service_external.functions.firstBlockAllowedToMonitor(
            closed_at_block=closed_at_block,
            settle_timeout=settle_timeout,
            participant1=to_checksum_address("0x%040x" % addresses[0]),
            participant2=to_checksum_address("0x%040x" % addresses[1]),
            monitoring_service_address=to_checksum_address("0x%040x" % addresses[2]),
        ).call()
        assert closed_at_block < first_allowed <= closed_at_block + settle_timeout
        return first_allowed

    # Basic example
    assert call([1, 2, 3]) == 1000 + 30 + (1 + 2 + 3)

    # Modulo used for one address
    assert call([100, 2, 3]) == 1000 + 30 + (100 % 50 + 2 + 3)

    # Modulo not used for any single address, but the sum of them
    assert call([40, 40, 40]) == 1000 + 30 + (40 + 40 + 40) % 50

    # Show that high address values don't cause overflows
    MAX_ADDRESS = 256 ** 20 - 1
    assert call([MAX_ADDRESS] * 3) == 1000 + 30 + (3 * MAX_ADDRESS) % 50

    # The highest settle_timeout does not cause overflows
    MAX_SETTLE_TIMEOUT = (2 ** 256 - 1) // 100 - 1
    assert call([1, 2, 3], settle_timeout=MAX_SETTLE_TIMEOUT) == 1000 + (
        MAX_SETTLE_TIMEOUT * 30
    ) // 100 + (1 + 2 + 3)

    # Extreme settle_timeout that would cause overflows will make the
    # transaction fail instead of giving the wrong result
    with pytest.raises(TransactionFailed, match="maliciously big settle timeout"):
        assert call([1, 2, 3], settle_timeout=MAX_SETTLE_TIMEOUT + 1)


def test_recoverAddressFromRewardProof(
    monitor_data_internal: Dict, token_network: Contract, monitoring_service_internals: Contract
) -> None:
    _, B = monitor_data_internal["participants"]

    recoverAddressFromRewardProof = (
        monitoring_service_internals.functions.recoverAddressFromRewardProofPublic
    )
    recovered_address = recoverAddressFromRewardProof(
        chain_id=token_network.functions.chain_id().call(),
        token_network_address=token_network.address,
        non_closing_participant=B,
        non_closing_signature=monitor_data_internal["non_closing_signature"],
        reward_amount=REWARD_AMOUNT,
        signature=monitor_data_internal["reward_proof_signature"],
    ).call()
    assert recovered_address == B


def test_monitoring_service_deploy_with_token_address_zero(
    deploy_tester_contract: Callable,
    service_registry: Contract,
    uninitialized_user_deposit_contract: Contract,
    token_network_registry_contract: Contract,
) -> None:
    # No good error message is available due to
    # https://github.com/raiden-network/raiden-contracts/issues/1329
    # with pytest.raises(TransactionFailed, match="Token at address zero"):
    with pytest.raises(TransactionFailed):
        deploy_tester_contract(
            contract_name=CONTRACT_MONITORING_SERVICE,
            _token_address=EMPTY_HEXADDRESS,  # causes error
            _service_registry_address=service_registry.address,
            _udc_address=uninitialized_user_deposit_contract.address,
            _token_network_registry_address=token_network_registry_contract.address,
        )


def test_monitoring_service_deploy_with_token_address_without_code(
    deploy_tester_contract: Callable,
    service_registry: Contract,
    uninitialized_user_deposit_contract: Contract,
    token_network_registry_contract: Contract,
) -> None:
    # No good error message is available due to
    # https://github.com/raiden-network/raiden-contracts/issues/1329
    # with pytest.raises(TransactionFailed, match="token has no code"):
    with pytest.raises(TransactionFailed):
        deploy_tester_contract(
            contract_name=CONTRACT_MONITORING_SERVICE,
            _token_address=DEPLOYER_ADDRESS,  # causes error
            _service_registry_address=service_registry.address,
            _udc_address=uninitialized_user_deposit_contract.address,
            _token_network_registry_address=token_network_registry_contract.address,
        )
    # See monitoring_service_external fixture for the success case.


def test_monitoring_service_deploy_with_service_registry_without_code(
    deploy_tester_contract: Callable,
    custom_token: Contract,
    uninitialized_user_deposit_contract: Contract,
    token_network_registry_contract: Contract,
) -> None:
    """Test MonitoringService's constructor with a service_registry address

    which does not contain code. This is supposed to cause a failure in one
    of the require() statements in the constructor."""
    # No good error message is available due to
    # https://github.com/raiden-network/raiden-contracts/issues/1329
    # with pytest.raises(TransactionFailed, match="ServiceRegistry has no code"):
    with pytest.raises(TransactionFailed):
        deploy_tester_contract(
            contract_name=CONTRACT_MONITORING_SERVICE,
            _token_address=custom_token.address,
            _service_registry_address=DEPLOYER_ADDRESS,  # causes error
            _udc_address=uninitialized_user_deposit_contract.address,
            _token_network_registry_address=token_network_registry_contract.address,
        )
    # See monitoring_service_external fixture for the success case.



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3.contract import Contract

from raiden_contracts.constants import MessageTypeId
from raiden_contracts.tests.utils import fake_bytes


def test_recover_address_from_withdraw_message(
    token_network_test_signatures: Contract,
    create_withdraw_signatures: Callable,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
) -> None:
    (A, B) = get_accounts(2)
    token_network = token_network_test_signatures
    fake_signature = fake_bytes(64)
    deposit_A = 5
    deposit_B = 7
    withdraw_A = 3
    expiration_block = 492889
    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    (signature_A, signature_B) = create_withdraw_signatures(
        [A, B],
        channel_identifier,
        A,
        withdraw_A,
        expiration_block,
        token_network.address,
    )

    recovered_address_A = token_network.functions.recoverAddressFromWithdrawMessagePublic(
        channel_identifier, A, withdraw_A, expiration_block, signature_A
    ).call()
    assert recovered_address_A == A

    recovered_address_B = token_network.functions.recoverAddressFromWithdrawMessagePublic(
        channel_identifier, A, withdraw_A, expiration_block, signature_B
    ).call()
    assert recovered_address_B == B

    with pytest.raises(TransactionFailed):
        token_network.functions.recoverAddressFromWithdrawMessagePublic(
            channel_identifier, A, withdraw_A, expiration_block, fake_signature
        ).call()

    wrong_participant = token_network.functions.recoverAddressFromWithdrawMessagePublic(
        channel_identifier, B, withdraw_A, expiration_block, signature_A
    ).call()
    assert recovered_address_A != wrong_participant

    wrong_withdraw_value = token_network.functions.recoverAddressFromWithdrawMessagePublic(
        channel_identifier, A, 1, expiration_block, signature_A
    ).call()

    assert recovered_address_A != wrong_withdraw_value

    wrong_signature = token_network.functions.recoverAddressFromWithdrawMessagePublic(
        channel_identifier, A, withdraw_A, expiration_block, signature_B
    ).call()

    assert recovered_address_A != wrong_signature

    wrong_expiration = token_network.functions.recoverAddressFromWithdrawMessagePublic(
        channel_identifier, A, withdraw_A, expiration_block + 3, signature_B
    ).call()

    assert recovered_address_A != wrong_expiration


def test_recover_address_from_balance_proof(
    token_network_test_signatures: Contract,
    create_balance_proof: Callable,
    get_accounts: Callable,
) -> None:
    """TokenNetwork can recover the signer's address from a balance proof

    This test checks that the TokenNetwork contract
    1) can recover the signer's address from a balance proof
    2) even when the signer and the balance proof's original sender are different
    3) recovers a wrong address when the balance proof is for a wrong token network
    """
    (A, B) = get_accounts(2)

    channel_identifier = 4
    balance_proof = create_balance_proof(
        channel_identifier, A, other_token_network=token_network_test_signatures
    )

    balance_proof_wrong_token_network = create_balance_proof(channel_identifier, A)

    balance_proof_other_signer = create_balance_proof(
        channel_identifier,
        A,
        signer=B,
        other_token_network=token_network_test_signatures,
    )

    assert (
        A
        == token_network_test_signatures.functions.recoverAddressFromBalanceProofPublic(
            channel_identifier, *balance_proof._asdict().values()
        ).call()
    )

    assert (
        B
        == token_network_test_signatures.functions.recoverAddressFromBalanceProofPublic(
            channel_identifier, *balance_proof_other_signer._asdict().values()
        ).call()
    )

    assert (
        A
        != token_network_test_signatures.functions.recoverAddressFromBalanceProofPublic(
            channel_identifier, *balance_proof_wrong_token_network._asdict().values()
        ).call()
    )


def test_recover_address_from_balance_proof_update(
    token_network_test_signatures: Contract,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    get_accounts: Callable,
) -> None:
    """TokenNetwork can recover the signer's address from a balance proof update

    This test checks that the TokenNetwork contract
    1) can recover the signer's address from a balance proof update
    2) recovers a wrong address if the balance proof update is for a wrong token network
    3) recovers a wrong address if the balance proof update is signed by the same secret key twice
    """
    (A, B) = get_accounts(2)
    other_token_network = token_network_test_signatures

    channel_identifier = 4
    balance_proof = create_balance_proof(
        channel_identifier, A, other_token_network=other_token_network
    )
    balance_proof_update_signature = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof._asdict(),
        other_token_network=other_token_network,
    )

    sig_wrong_token_network = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof._asdict(),
    )

    balance_proof_signed_B = create_balance_proof(
        channel_identifier, B, other_token_network=other_token_network
    )
    balance_proof_update_signature_wrong_signer = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_signed_B._asdict(),
    )

    assert (
        B
        == other_token_network.functions.recoverAddressFromBalanceProofCounterSignaturePublic(
            MessageTypeId.BALANCE_PROOF_UPDATE,
            channel_identifier,
            *balance_proof._asdict().values(),
            balance_proof_update_signature,
        ).call()
    )

    assert (
        B
        != other_token_network.functions.recoverAddressFromBalanceProofCounterSignaturePublic(
            MessageTypeId.BALANCE_PROOF_UPDATE,
            channel_identifier,
            *balance_proof._asdict().values(),
            sig_wrong_token_network,
        ).call()
    )

    assert (
        B
        != other_token_network.functions.recoverAddressFromBalanceProofCounterSignaturePublic(
            MessageTypeId.BALANCE_PROOF_UPDATE,
            channel_identifier,
            *balance_proof._asdict().values(),
            balance_proof_update_signature_wrong_signer,
        ).call()
    )


def test_recover_address_from_balance_proof_close(
    token_network_test_signatures: Contract,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    get_accounts: Callable,
) -> None:
    """TokenNetwork can recover the signer's address from a closing proof

    This test checks that the TokenNetwork contract
    1) can recover the signer's address from a balance proof update
    2) recovers a wrong address if the balance proof update is for a wrong token network
    3) recovers a wrong address if the balance proof update is signed by the same secret key twice
    """
    (A, B) = get_accounts(2)
    other_token_network = token_network_test_signatures

    channel_identifier = 4
    balance_proof = create_balance_proof(
        channel_identifier, A, other_token_network=other_token_network
    )
    balance_proof_update_signature = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof._asdict(),
        other_token_network=other_token_network,
    )

    sig_wrong_token_network = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof._asdict(),
    )

    balance_proof_signed_B = create_balance_proof(
        channel_identifier, B, other_token_network=other_token_network
    )
    balance_proof_update_signature_wrong_signer = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_signed_B._asdict(),
    )

    assert (
        B
        == other_token_network.functions.recoverAddressFromBalanceProofCounterSignaturePublic(
            MessageTypeId.BALANCE_PROOF,
            channel_identifier,
            *balance_proof._asdict().values(),
            balance_proof_update_signature,
        ).call()
    )

    assert (
        B
        != other_token_network.functions.recoverAddressFromBalanceProofCounterSignaturePublic(
            MessageTypeId.BALANCE_PROOF,
            channel_identifier,
            *balance_proof._asdict().values(),
            sig_wrong_token_network,
        ).call()
    )

    assert (
        B
        != other_token_network.functions.recoverAddressFromBalanceProofCounterSignaturePublic(
            MessageTypeId.BALANCE_PROOF,
            channel_identifier,
            *balance_proof._asdict().values(),
            balance_proof_update_signature_wrong_signer,
        ).call()
    )



from collections import namedtuple
from copy import deepcopy
from typing import Tuple

from eth_typing import HexAddress
from eth_utils import encode_hex, keccak, to_canonical_address
from web3.contract import Contract

from raiden_contracts.constants import LOCKSROOT_OF_NO_LOCKS
from raiden_contracts.tests.utils.constants import EMPTY_ADDITIONAL_HASH, UINT256_MAX

SettlementValues = namedtuple(
    "SettlementValues",
    [
        "participant1_balance",
        "participant2_balance",
        "participant1_locked",
        "participant2_locked",
    ],
)


class LockedAmounts:
    def __init__(self, claimable_locked: int = 0, unclaimable_locked: int = 0):
        self.claimable_locked = claimable_locked
        self.unclaimable_locked = unclaimable_locked

    @property
    def locked(self) -> int:
        return self.claimable_locked + self.unclaimable_locked


ZERO_LOCKED_VALUES = LockedAmounts()


class ChannelValues:
    def __init__(
        self,
        deposit: int = 0,
        withdrawn: int = 0,
        nonce: int = 0,
        transferred: int = 0,
        locked_amounts: LockedAmounts = ZERO_LOCKED_VALUES,
        locksroot: bytes = LOCKSROOT_OF_NO_LOCKS,
        additional_hash: bytes = EMPTY_ADDITIONAL_HASH,
    ):
        self.deposit = deposit
        self.withdrawn = withdrawn
        self.nonce = nonce
        self.transferred = transferred
        # deepcopy is necessary because later some elements are incremented in-place.
        self.locked_amounts = deepcopy(locked_amounts)
        self.locksroot = locksroot
        self.additional_hash = additional_hash

    def __repr__(self) -> str:
        return (
            f"ChannelValues deposit:{self.deposit} withdrawn:{self.withdrawn} "
            f"transferred:{self.transferred} claimable_"
            f"locked:{self.locked_amounts.claimable_locked} "
            f"unclaimable_locked:{self.locked_amounts.unclaimable_locked} "
            f"locked:{self.locked_amounts.locked} "
            f"locksroot:{encode_hex(self.locksroot)}"
        )


def get_participant_available_balance(
    participant1: ChannelValues, participant2: ChannelValues
) -> int:
    """Returns the available balance for participant1

    The available balance is used in the Raiden client, in order to check if a
    participant is able to make transfers or not.
    """
    return (
        participant1.deposit
        + participant2.transferred
        - participant1.withdrawn
        - participant1.transferred
        - participant1.locked_amounts.locked
    )


def are_balance_proofs_valid(participant1: ChannelValues, participant2: ChannelValues) -> bool:
    """Checks if balance proofs are valid or could have been valid at a certain point in time

    Balance proof constraints are detailed in
    https://github.com/raiden-network/raiden-contracts/issues/188#issuecomment-404752095
    """
    participant1_available_balance = get_participant_available_balance(participant1, participant2)
    participant2_available_balance = get_participant_available_balance(participant2, participant1)

    total_available_deposit = get_total_available_deposit(participant1, participant2)

    return (
        0 <= participant1.transferred <= UINT256_MAX
        and 0 <= participant2.transferred <= UINT256_MAX
        and 0 <= participant1.locked_amounts.locked <= UINT256_MAX
        and 0 <= participant2.locked_amounts.locked <= UINT256_MAX
        and participant1.transferred + participant1.locked_amounts.locked <= UINT256_MAX
        and participant2.transferred + participant2.locked_amounts.locked <= UINT256_MAX
        and 0 <= participant1_available_balance <= total_available_deposit
        and 0 <= participant2_available_balance <= total_available_deposit
        and total_available_deposit <= UINT256_MAX
        and participant1.locked_amounts.locked <= participant1_available_balance
        and participant2.locked_amounts.locked <= participant2_available_balance
    )


def were_balance_proofs_valid(participant1: ChannelValues, participant2: ChannelValues) -> bool:
    """Checks if balance proofs were ever valid."""
    deposit = participant1.deposit + participant2.deposit

    # Regardless of issuance time, the locked amount must be smaller than the
    # total channel deposit.
    return (
        participant1.locked_amounts.locked <= deposit
        and participant2.locked_amounts.locked <= deposit
    )


def is_balance_proof_old(participant1: ChannelValues, participant2: ChannelValues) -> bool:
    """Checks if balance proofs are valid, with at least one old."""
    assert were_balance_proofs_valid(participant1, participant2)

    total_available_deposit = get_total_available_deposit(participant1, participant2)

    (
        participant1_balance,
        participant2_balance,
    ) = get_expected_after_settlement_unlock_amounts(participant1, participant2)

    # Valid last balance proofs should ensure the following equality:
    if participant1_balance + participant2_balance == total_available_deposit:
        return False

    return True


def get_settlement_amounts(
    participant1: ChannelValues, participant2: ChannelValues
) -> SettlementValues:
    """Settlement algorithm

    Calculates the token amounts to be transferred to the channel participants when
    a channel is settled.

    !!! Don't change this unless you really know what you are doing.
    """
    total_available_deposit = get_total_available_deposit(participant1, participant2)
    participant1_max_transferred = min(
        participant1.transferred + participant1.locked_amounts.locked, UINT256_MAX
    )
    participant2_max_transferred = min(
        participant2.transferred + participant2.locked_amounts.locked, UINT256_MAX
    )
    participant1_max_amount_receivable = (
        participant1.deposit
        + participant2_max_transferred
        - participant1_max_transferred
        - participant1.withdrawn
    )

    participant1_max_amount_receivable = max(participant1_max_amount_receivable, 0)
    participant1_max_amount_receivable = min(
        participant1_max_amount_receivable, total_available_deposit
    )
    participant2_max_amount_receivable = (
        total_available_deposit - participant1_max_amount_receivable
    )

    participant2_locked = min(
        participant2.locked_amounts.locked, participant1_max_amount_receivable
    )
    participant1_locked = min(
        participant1.locked_amounts.locked, participant2_max_amount_receivable
    )

    participant1_amount = participant1_max_amount_receivable - participant2_locked
    participant2_amount = participant2_max_amount_receivable - participant1_locked

    assert total_available_deposit == (
        participant1_amount + participant2_amount + participant1_locked + participant2_locked
    )

    return SettlementValues(
        participant1_balance=participant1_amount,
        participant2_balance=participant2_amount,
        participant1_locked=participant1_locked,
        participant2_locked=participant2_locked,
    )


def get_expected_after_settlement_unlock_amounts(
    participant1: ChannelValues, participant2: ChannelValues
) -> Tuple[int, int]:
    """Get expected balances after the channel is settled and all locks are unlocked

    We make the assumption that both balance proofs provided are valid, meaning that both
    participants' balance proofs are the last known balance proofs from the channel.
    """
    participant1_balance = (
        participant1.deposit
        - participant1.withdrawn
        + participant2.transferred
        - participant1.transferred
        + participant2.locked_amounts.claimable_locked
        - participant1.locked_amounts.claimable_locked
    )
    participant2_balance = (
        participant2.deposit
        - participant2.withdrawn
        + participant1.transferred
        - participant2.transferred
        + participant1.locked_amounts.claimable_locked
        - participant2.locked_amounts.claimable_locked
    )
    return (participant1_balance, participant2_balance)


def failsafe_add(a: int, b: int) -> int:
    """
    Function to compute the settlement amounts w.r.t. MAX_UINT, as it happens in the contract
    :param a: Addend
    :param b: Addend
    :return: sum, if a+b mod MAX_UINT+1 would not overflow else: MAX_UINT256
    """
    a = a % (UINT256_MAX + 1)
    b = b % (UINT256_MAX + 1)
    ret = (a + b) % (UINT256_MAX + 1)
    if ret >= a:
        return ret
    else:
        return UINT256_MAX


def failsafe_sub(a: int, b: int) -> Tuple[int, int]:
    """
    Function to compute the settlement amounts w.r.t. MAX_UINT, as it happens in the contract
    :param a: Minuend
    :param b: Subtrahend
    :return: tuple(difference, Subtrahend) if a-b mod MAX_UINT+1 would not underflow
     else: tuple(0, Minuend)
    """
    a = a % (UINT256_MAX + 1)
    b = b % (UINT256_MAX + 1)
    return (a - b, b) if a > b else (0, a)


def get_onchain_settlement_amounts(
    participant1: ChannelValues, participant2: ChannelValues
) -> SettlementValues:
    """Settlement algorithm

    Calculates the token amounts to be transferred to the channel participants when
    a channel is settled.

    !!! Don't change this unless you really know what you are doing.
    """

    assert (
        participant2.transferred + participant2.locked_amounts.locked
        >= participant1.transferred + participant1.locked_amounts.locked
    )

    total_available_deposit = get_total_available_deposit(participant1, participant2)

    # we assume that total_available_deposit does not overflow in settleChannel
    assert total_available_deposit <= UINT256_MAX

    participant1_max_transferred = failsafe_add(
        participant1.transferred, participant1.locked_amounts.locked
    )
    participant2_max_transferred = failsafe_add(
        participant2.transferred, participant2.locked_amounts.locked
    )

    assert participant1_max_transferred <= UINT256_MAX
    assert participant2_max_transferred <= UINT256_MAX

    participant1_net_max_transferred = participant2_max_transferred - participant1_max_transferred

    participant1_max_amount = failsafe_add(participant1_net_max_transferred, participant1.deposit)
    (participant1_max_amount, _) = failsafe_sub(participant1_max_amount, participant1.withdrawn)

    participant1_max_amount = min(participant1_max_amount, total_available_deposit)

    participant2_max_amount = total_available_deposit - participant1_max_amount

    (participant1_amount, participant2_locked_amount) = failsafe_sub(
        participant1_max_amount, participant2.locked_amounts.locked
    )

    (participant2_amount, participant1_locked_amount) = failsafe_sub(
        participant2_max_amount, participant1.locked_amounts.locked
    )

    assert total_available_deposit == (
        participant1_amount
        + participant2_amount
        + participant1_locked_amount
        + participant2_locked_amount
    )

    return SettlementValues(
        participant1_balance=participant1_amount,
        participant2_balance=participant2_amount,
        participant1_locked=participant1_locked_amount,
        participant2_locked=participant2_locked_amount,
    )


def get_total_available_deposit(participant1: ChannelValues, participant2: ChannelValues) -> int:
    total_available_deposit = (
        participant1.deposit
        + participant2.deposit
        - participant1.withdrawn
        - participant2.withdrawn
    )
    return total_available_deposit


def get_unlocked_amount(secret_registry: Contract, packed_locks: bytes) -> int:
    unlocked_amount = 0

    for i in range(0, len(packed_locks), 96):
        lock = packed_locks[i : (i + 96)]
        expiration_block = int.from_bytes(lock[0:32], byteorder="big")
        locked_amount = int.from_bytes(lock[32:64], byteorder="big")
        secrethash = lock[64:96]

        reveal_block = secret_registry.functions.getSecretRevealBlockHeight(secrethash).call()
        if 0 < reveal_block < expiration_block:
            unlocked_amount += locked_amount
    return unlocked_amount


def get_participants_hash(A: HexAddress, B: HexAddress) -> bytes:
    A_canonical = to_canonical_address(A)
    B_canonical = to_canonical_address(B)
    if A_canonical == B_canonical:
        raise ValueError("get_participants_hash got the same address twice")
    return (
        keccak(A_canonical + B_canonical)
        if A_canonical < B_canonical
        else keccak(B_canonical + A_canonical)
    )



from tempfile import NamedTemporaryFile

import pytest

from raiden_contracts.utils.join_contracts import ContractJoiner


def test_contract_joiner_with_non_existent_import() -> None:
    """Using ContractJoiner on a source importing a nonexistent file"""
    joiner = ContractJoiner()
    source_with_missing_import = """
        import "NonExistent.sol";
    """
    with NamedTemporaryFile() as source_file:
        source_file.write(bytearray(source_with_missing_import, "ascii"))
        source_file.flush()
        with pytest.raises(FileNotFoundError):
            joiner.join(open(source_file.name))


def test_contract_joiner_with_empty_file() -> None:
    """Using ContractJoiner on an empty source"""
    joiner = ContractJoiner()
    with NamedTemporaryFile() as source_file:
        assert [] == joiner.join(open(source_file.name))



"""ContractSourceManager knows the sources and how to compile them."""
import hashlib
import json
from os import chdir
from pathlib import Path
from typing import Dict, Optional, Tuple

import solcx

from raiden_contracts.constants import PRECOMPILED_DATA_FIELDS, DeploymentModule
from raiden_contracts.contract_manager import ContractManager, contracts_data_path

_BASE = Path(__file__).parent
SOLC_VERSION = "0.7.6"


class ContractSourceManagerCompilationError(RuntimeError):
    """Compilation failed for infrastructural reasons (lack of the compiler,
    failure to take checksums)."""


class ContractSourceManagerVerificationError(RuntimeError):
    """Failure in comparing contracts.json contents against sources."""


class ContractSourceManager:
    """ContractSourceManager knows how to compile contracts"""

    def __init__(self, path: Dict[str, Path]) -> None:
        """Params: path: a dictionary of directories which contain solidity files to compile"""
        if not isinstance(path, dict):
            raise TypeError("Wrong type of argument given for ContractSourceManager()")
        self.contracts_source_dirs = path
        (self.contracts_checksums, self.overall_checksum) = self._checksum_contracts()

    def _compile_all_contracts(self) -> Dict:
        """
        Compile solidity contracts into ABI and BIN. This requires solc somewhere in the $PATH
        and also the :ref:`ethereum.tools` python library.  The return value is a dict that
        should be written into contracts.json.
        """
        solcx.install.install_solc(SOLC_VERSION)
        solcx.set_solc_version(SOLC_VERSION)
        ret = {}
        old_working_dir = Path.cwd()
        chdir(_BASE)

        def relativise(path: Path) -> Path:
            return path.relative_to(_BASE)

        import_dir_map = [
            "%s=%s" % (k, relativise(v)) for k, v in self.contracts_source_dirs.items()
        ]
        import_dir_map.insert(0, ".=.")  # allow solc to compile contracts in all subdirs
        try:
            for contracts_dir in self.contracts_source_dirs.values():
                res = solcx.compile_files(
                    [str(relativise(file)) for file in contracts_dir.glob("*.sol")],
                    output_values=PRECOMPILED_DATA_FIELDS,
                    import_remappings=import_dir_map,
                    optimize=True,
                    optimize_runs=200,
                )

                ret.update(_fix_contract_key_names(res))
        except FileNotFoundError as ex:
            raise ContractSourceManagerCompilationError(
                "Could not compile the contract. Check that solc is available."
            ) from ex
        finally:
            chdir(old_working_dir)
        check_runtime_codesize(ret)
        return ret

    def compile_contracts(self, target_path: Path) -> ContractManager:
        """Store compiled contracts JSON at `target_path`."""
        assert self.overall_checksum is not None

        contracts_compiled = self._compile_all_contracts()

        target_path.parent.mkdir(parents=True, exist_ok=True)
        with target_path.open(mode="w") as target_file:
            target_file.write(
                json.dumps(
                    dict(
                        contracts=contracts_compiled,
                        contracts_checksums=self.contracts_checksums,
                        overall_checksum=self.overall_checksum,
                        contracts_version=None,
                    ),
                    sort_keys=True,
                    indent=4,
                )
            )

        return ContractManager(target_path)

    def verify_precompiled_checksums(self, precompiled_path: Path) -> None:
        """Compare source code checksums with those from a precompiled file

        If `contract_name` is None, all contracts checksums and the overall checksum are checked.
        """

        # We get the precompiled file data
        contracts_precompiled = ContractManager(precompiled_path)

        # Compare each contract source code checksum with the one from the precompiled file
        for contract, checksum in self.contracts_checksums.items():
            _verify_single_precompiled_checksum(
                checked_checksums=contracts_precompiled.contracts_checksums,
                contract_name=contract,
                expected_checksum=checksum,
            )

        # Compare the overall source code checksum with the one from the precompiled file
        if self.overall_checksum != contracts_precompiled.overall_checksum:
            raise ContractSourceManagerVerificationError(
                f"overall checksum does not match "
                f"{self.overall_checksum} != {contracts_precompiled.overall_checksum}"
            )

    def _checksum_contracts(self) -> Tuple[Dict[str, str], str]:
        """Compute the checksum of each source, and the overall checksum

        Returns (contracts_checksums, overall_checksum)
        """
        checksums: Dict[str, str] = {}
        for contracts_dir in self.contracts_source_dirs.values():
            file: Path
            for file in contracts_dir.glob("*.sol"):
                checksums[file.name] = hashlib.sha256(file.read_bytes()).hexdigest()

        overall_checksum = hashlib.sha256(
            ":".join(checksums[key] for key in sorted(checksums)).encode()
        ).hexdigest()
        return (checksums, overall_checksum)


def contracts_source_path(contracts_version: Optional[str]) -> Dict[str, Path]:
    data = contracts_data_path(contracts_version)
    return contracts_source_path_with_stem(data.joinpath("source"))


def contracts_source_path_of_deployment_module(module: DeploymentModule) -> Path:
    if module == DeploymentModule.RAIDEN:
        return contracts_source_path(contracts_version=None)["raiden"]
    elif module == DeploymentModule.SERVICES:
        return contracts_source_path(contracts_version=None)["services"]
    else:
        raise ValueError(f"No source known for module {module}")


def contracts_source_path_with_stem(stem: Path) -> Dict[str, Path]:
    """The directory remapping given to the Solidity compiler."""
    return {
        "lib": _BASE.joinpath(stem, "lib"),
        "raiden": _BASE.joinpath(stem, "raiden"),
        "test": _BASE.joinpath(stem, "test"),
        "services": _BASE.joinpath(stem, "services"),
    }


def _fix_contract_key_names(d: Dict) -> Dict:
    result = {}

    for k, v in d.items():
        name = k.split(":")[1]
        result[name] = v

    return result


def check_runtime_codesize(d: Dict) -> None:
    """Raises a RuntimeError if the runtime codesize exceeds the EIP-170 limit"""
    for name, compilation in d.items():
        runtime_code_len = len(compilation["bin-runtime"]) // 2
        if runtime_code_len > 0x6000:
            raise RuntimeError(f"{name}'s runtime code is too big ({runtime_code_len} bytes).")


def _verify_single_precompiled_checksum(
    checked_checksums: Dict[str, str], contract_name: str, expected_checksum: str
) -> None:
    """Get `checked_checksums[contract_name]` and compare it against `expected_checksum`"""
    try:
        precompiled_checksum = checked_checksums[contract_name]
    except KeyError:
        raise ContractSourceManagerVerificationError(f"No checksum for {contract_name}")
    if precompiled_checksum != expected_checksum:
        raise ContractSourceManagerVerificationError(
            f"checksum of {contract_name} does not match. got {precompiled_checksum} != "
            f"expected {expected_checksum}"
        )


def verify_single_precompiled_checksum_on_nonexistent_contract_name() -> None:
    """A functiohn for testing the case where the contract name is not found"""
    _verify_single_precompiled_checksum(
        checked_checksums={}, contract_name="a", expected_checksum="abc"
    )



from collections import namedtuple
from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    EMPTY_ADDRESS,
    TEST_SETTLE_TIMEOUT_MIN,
    ChannelEvent,
    ChannelState,
    MessageTypeId,
)
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    ChannelValues,
    call_and_transact,
    fake_bytes,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.events import check_transfer_updated


def test_update_call(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Call updateNonClosingBalanceProof() with various wrong arguments"""
    (A, B, C) = get_accounts(3)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 15, B)
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    # Failure with the zero address instead of A's address
    with pytest.raises(TransactionFailed, match="TN: participant address zero"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            EMPTY_ADDRESS,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": C})

    # Failure with the zero address instead of B's address
    with pytest.raises(TransactionFailed, match="TN: partner address zero"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            EMPTY_ADDRESS,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": C})

    # Failure with the zero signature
    with pytest.raises(TransactionFailed):  # TODO: sig check
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            EMPTY_SIGNATURE,
        ).call({"from": C})

    # Failure with the empty balance hash
    with pytest.raises(TransactionFailed, match="TN/update: balance hash is zero"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            EMPTY_BALANCE_HASH,
            balance_proof_A.nonce,
            balance_proof_A.additional_hash,
            balance_proof_A.original_signature,
            balance_proof_update_signature_B,
        ).call({"from": C})

    # Failure with nonce zero
    with pytest.raises(TransactionFailed, match="TN/update: nonce is zero"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_A.balance_hash,
            0,
            balance_proof_A.additional_hash,
            balance_proof_A.original_signature,
            balance_proof_update_signature_B,
        ).call({"from": C})

    # Failure with the empty signature
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_A.balance_hash,
            balance_proof_A.nonce,
            balance_proof_A.additional_hash,
            EMPTY_SIGNATURE,
            balance_proof_update_signature_B,
        ).call({"from": C})

    # See a success to make sure the above failures are not spurious
    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_A.balance_hash,
            balance_proof_A.nonce,
            balance_proof_A.additional_hash,
            balance_proof_A.original_signature,
            balance_proof_update_signature_B,
        ),
        {"from": C},
    )


def test_update_nonexistent_fail(
    get_accounts: Callable,
    token_network: Contract,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """updateNonClosingBalanceProof() on a not-yet openned channel should fail"""
    (A, B, C) = get_accounts(3)
    channel_identifier = 1

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert settle_block_number == 0
    assert state == ChannelState.NONEXISTENT

    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    with pytest.raises(TransactionFailed, match="TN/update: channel id mismatch"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": C})


def test_update_notclosed_fail(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """updateNonClosingBalanceProof() on an Opened channel should fail"""
    (A, B, C) = get_accounts(3)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 25, B)

    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert settle_block_number > 0
    assert state == ChannelState.OPENED

    with pytest.raises(TransactionFailed, match="TN/update: channel not closed"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": C})


def test_update_wrong_nonce_fail(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    update_state_tests: Callable,
) -> None:
    (A, B, Delegate) = get_accounts(3)
    settle_timeout = 6
    deposit_A = 20
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)
    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, 0, 3, fake_bytes(32, "02"))
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )
    txn_hash1 = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )

    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": Delegate},
    )

    # updateNonClosingBalanceProof should fail for the same nonce as provided previously
    with pytest.raises(TransactionFailed, match="TN: nonce reused"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": Delegate})
    balance_proof_A_same_nonce = create_balance_proof(
        channel_identifier, A, 12, 2, balance_proof_A.nonce, fake_bytes(32, "03")
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A_same_nonce._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": Delegate})

    balance_proof_A_lower_nonce = create_balance_proof(
        channel_identifier, A, 10, 0, 4, fake_bytes(32, "02")
    )

    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A_lower_nonce._asdict(),
    )
    with pytest.raises(TransactionFailed, match="TN: nonce reused"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A_lower_nonce._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": A})

    update_state_tests(
        channel_identifier,
        A,
        balance_proof_A,
        B,
        balance_proof_B,
        settle_timeout,
        txn_hash1,
    )


def test_update_wrong_signatures(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """updateNonClosingBalanceProof() should fail with wrong signatures"""
    (A, B, C) = get_accounts(3)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 25, B)

    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_A_fake = create_balance_proof(
        channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"), signer=C
    )

    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )
    balance_proof_update_signature_B_fake = create_balance_proof_countersignature(
        participant=C,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    # Close the channel so updateNonClosingBalanceProof() is possible
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A_fake._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": C})
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B_fake,
        ).call({"from": C})

    # See a success to make sure that the above failures are not spurious
    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": C},
    )


def test_update_channel_state(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    update_state_tests: Callable,
    txn_cost: Callable,
) -> None:
    """A successful updateNonClosingBalanceProof() call should not change token/ETH balances"""
    (A, B, Delegate) = get_accounts(3)
    settle_timeout = 6
    deposit_A = 20
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)
    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, 0, 3, fake_bytes(32, "02"))
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    txn_hash1 = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )

    pre_eth_balance_A = web3.eth.get_balance(A)
    pre_eth_balance_B = web3.eth.get_balance(B)
    pre_eth_balance_delegate = web3.eth.get_balance(Delegate)
    pre_eth_balance_contract = web3.eth.get_balance(token_network.address)
    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_delegate = custom_token.functions.balanceOf(Delegate).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    txn_hash = call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": Delegate},
    )

    # Test that no balances have changed.
    # There are no transfers to be made in updateNonClosingBalanceProof.
    assert web3.eth.get_balance(A) == pre_eth_balance_A
    assert web3.eth.get_balance(B) == pre_eth_balance_B
    assert web3.eth.get_balance(Delegate) == pre_eth_balance_delegate - txn_cost(txn_hash)
    assert web3.eth.get_balance(token_network.address) == pre_eth_balance_contract
    assert custom_token.functions.balanceOf(A).call() == pre_balance_A
    assert custom_token.functions.balanceOf(B).call() == pre_balance_B
    assert custom_token.functions.balanceOf(Delegate).call() == pre_balance_delegate
    assert custom_token.functions.balanceOf(token_network.address).call() == pre_balance_contract

    update_state_tests(
        channel_identifier,
        A,
        balance_proof_A,
        B,
        balance_proof_B,
        settle_timeout,
        txn_hash1,
    )


def test_update_channel_fail_no_offchain_transfers(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Calls to updateNonClosingBalanceProof() fail with the zero nonce"""
    (A, B) = get_accounts(2)

    channel_identifier = create_channel(A, B)[0]
    balance_proof_A = create_balance_proof(channel_identifier, A, 0, 0, 0)
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    with pytest.raises(TransactionFailed, match="TN/update: balance hash is zero"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            EMPTY_SIGNATURE,
        ).call({"from": B})

    with pytest.raises(TransactionFailed, match="TN/update: nonce is zero"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": B})


def test_update_allowed_after_settlement_period(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    web3: Web3,
) -> None:
    """updateNonClosingBalanceProof can be called after the settlement period."""
    (A, B) = get_accounts(2)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    deposit_A = 20
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)
    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 5, fake_bytes(32, "02"))
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, 0, 3, fake_bytes(32, "02"))
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )
    mine_blocks(web3, settle_timeout + 1)
    token_network.functions.updateNonClosingBalanceProof(
        channel_identifier,
        A,
        B,
        *balance_proof_A._asdict().values(),
        balance_proof_update_signature_B,
    ).call({"from": A})


def test_update_not_allowed_for_the_closing_address(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """Closing address cannot call updateNonClosingBalanceProof."""
    (A, B, M) = get_accounts(3)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    deposit_A = 20
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    # Some balance proof from B
    balance_proof_B_0 = create_balance_proof(channel_identifier, B, 5, 0, 3, fake_bytes(32, "02"))
    closing_sig_A_0 = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B_0._asdict(),
    )

    # Later balance proof, higher transferred amount, higher nonce
    balance_proof_B_1 = create_balance_proof(channel_identifier, B, 10, 0, 4, fake_bytes(32, "02"))

    # B's signature on the update message is valid
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_B_1._asdict(),
    )

    # A closes with the first balance proof
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            *balance_proof_B_0._asdict().values(),
            closing_sig_A_0,
        ),
        {"from": A},
    )

    # Someone wants to update with later balance proof - not possible
    with pytest.raises(TransactionFailed, match="TN/update: invalid closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_B_1._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/update: invalid closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_B_1._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": B})
    with pytest.raises(TransactionFailed, match="TN/update: invalid closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_B_1._asdict().values(),
            balance_proof_update_signature_B,
        ).call({"from": M})


def test_update_invalid_balance_proof_arguments(
    token_network: Contract,
    token_network_test_utils: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """updateNonClosingBalanceProof() should fail on balance proofs with various wrong params"""
    (A, B, C) = get_accounts(3)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    deposit_A = 20
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    balance_proof = namedtuple(
        "balance_proof", ["balance_hash", "nonce", "additional_hash", "signature"]
    )

    #  Create valid balance_proof
    balance_proof_valid = balance_proof(
        *create_balance_proof(channel_identifier, A, 10, 0, 2, fake_bytes(32, "02"))
        ._asdict()
        .values()
    )

    # And a valid nonclosing_signature
    valid_balance_proof_update_signature = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        *balance_proof_valid._asdict().values(),
    )

    #  We test invalid balance proof arguments with valid signatures

    #  Create balance_proof for invalid token_network
    balance_proof_invalid_token_network = balance_proof(
        *create_balance_proof(
            channel_identifier,
            A,
            10,
            0,
            2,
            fake_bytes(32, "02"),
            other_token_network=token_network_test_utils,
        )
        ._asdict()
        .values()
    )

    signature_invalid_token_network = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
        other_token_network=token_network_test_utils,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_invalid_token_network._asdict().values(),
            signature_invalid_token_network,
        ).call({"from": B})

    #  Create balance_proof for invalid channel participant
    balance_proof_invalid_channel_participant = balance_proof(
        *create_balance_proof(channel_identifier, C, 10, 0, 2, fake_bytes(32, "02"))
        ._asdict()
        .values()
    )

    signature_invalid_channel_participant = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_invalid_channel_participant._asdict().values(),
            signature_invalid_channel_participant,
        ).call({"from": B})

    #  Create balance_proof for invalid channel identifier
    balance_proof_invalid_channel_identifier = balance_proof(
        *create_balance_proof(channel_identifier + 1, A, 10, 0, 2, fake_bytes(32, "02"))
        ._asdict()
        .values()
    )

    signature_invalid_channel_identifier = create_balance_proof_countersignature(
        B,
        channel_identifier + 1,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_invalid_channel_identifier._asdict().values(),
            signature_invalid_channel_identifier,
        ).call({"from": B})

    signature_invalid_balance_hash = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash[::-1],
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_valid.balance_hash[::-1],  # invalid balance_hash
            balance_proof_valid.nonce,
            balance_proof_valid.additional_hash,
            balance_proof_valid.signature,
            signature_invalid_balance_hash,
        ).call({"from": B})

    signature_invalid_nonce = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        1,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )

    with pytest.raises(TransactionFailed, match="TN/update: invalid closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_valid.balance_hash,
            1,  # invalid nonce
            balance_proof_valid.additional_hash,
            balance_proof_valid.signature,
            signature_invalid_nonce,
        ).call({"from": B})

    signature_invalid_additional_hash = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash[::-1],
        balance_proof_valid.signature,
    )

    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_valid.balance_hash,
            balance_proof_valid.nonce,
            fake_bytes(32, "02"),  # invalid additional_hash
            balance_proof_valid.signature,
            signature_invalid_additional_hash,
        ).call({"from": B})

    signature_invalid_closing_signature = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature[::-1],
    )

    with pytest.raises(TransactionFailed):  # TODO: fixme
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            balance_proof_valid.balance_hash,
            balance_proof_valid.nonce,
            balance_proof_valid.additional_hash,
            balance_proof_valid.signature,
            signature_invalid_closing_signature[::-1],  # invalid non-closing signature
        ).call({"from": B})

    # Call with same balance_proof and signature on valid arguments still works

    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            valid_balance_proof_update_signature,
        ),
        {"from": B},
    )


def test_update_signature_on_invalid_arguments(
    token_network: Contract,
    token_network_test_utils: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Call updateNonClosingBalanceProof with signature on invalid argument fails"""
    (A, B, C) = get_accounts(3)
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    deposit_A = 20
    channel_identifier = create_channel(A, B, settle_timeout)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)
    balance_proof = namedtuple(
        "balance_proof", ["balance_hash", "nonce", "additional_hash", "signature"]
    )

    # Close channel
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    #  Create valid balance_proof
    balance_proof_valid = balance_proof(
        *create_balance_proof(
            channel_identifier, A, 10, 0, 2, fake_bytes(32, "02"), fake_bytes(32, "02")
        )
        ._asdict()
        .values()
    )

    signature_invalid_token_network_address = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        *balance_proof_valid._asdict().values(),
        other_token_network=token_network_test_utils,  # invalid token_network_address
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_token_network_address,
        ).call({"from": B})

    signature_invalid_participant = create_balance_proof_countersignature(
        C,  # invalid signer
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_participant,
        ).call({"from": B})

    signature_invalid_channel_identifier = create_balance_proof_countersignature(
        B,
        channel_identifier + 1,  # invalid channel_identifier
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_channel_identifier,
        ).call({"from": B})

    signature_invalid_balance_hash = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash[::-1],  # invalid balance_hash
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_balance_hash,
        ).call({"from": B})

    signature_invalid_nonce = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        1,  # invalid nonce
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_nonce,
        ).call({"from": B})

    signature_invalid_additional_hash = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        b"\x00" * 32,  # invalid additional_hash
        balance_proof_valid.signature,
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_additional_hash,
        ).call({"from": B})

    signature_invalid_closing_signature = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        balance_proof_valid.balance_hash,
        balance_proof_valid.nonce,
        balance_proof_valid.additional_hash,
        balance_proof_valid.signature[::-1],
    )
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            signature_invalid_closing_signature,
        ).call({"from": B})

    # Call with same balance_proof and signature on valid arguments works
    balance_proof_update_signature = create_balance_proof_countersignature(
        B,
        channel_identifier,
        MessageTypeId.BALANCE_PROOF_UPDATE,
        *balance_proof_valid._asdict().values(),
    )
    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_valid._asdict().values(),
            balance_proof_update_signature,
        ),
        {"from": B},
    )


def test_update_replay_reopened_channel(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """updateNonClosingBalanceProof() should refuse a balance proof with a stale channel id"""
    (A, B) = get_accounts(2)
    nonce_B = 5
    values_A = ChannelValues(deposit=10, transferred=0)
    values_B = ChannelValues(deposit=20, transferred=15)

    channel_identifier1 = create_channel(A, B)[0]
    channel_deposit(channel_identifier1, B, values_B.deposit, A)
    balance_proof_B = create_balance_proof(
        channel_identifier1,
        B,
        values_B.transferred,
        values_B.locked_amounts.locked,
        nonce_B,
        values_B.locksroot,
    )
    balance_proof_update_signature_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier1,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_B._asdict(),
    )

    closing_sig = create_close_signature_for_no_balance_proof(B, channel_identifier1)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier1,
            A,
            B,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": B},
    )

    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier1,
            B,
            A,
            *balance_proof_B._asdict().values(),
            balance_proof_update_signature_A,
        ),
        {"from": A},
    )

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier1,
            A,
            values_A.transferred,
            values_A.locked_amounts.locked,
            values_A.locksroot,
            B,
            values_B.transferred,
            values_B.locked_amounts.locked,
            values_B.locksroot,
        ),
        {"from": A},
    )

    # Make sure we cannot update balance proofs after settleChannel is called
    with pytest.raises(TransactionFailed, match="TN/update: channel id mismatch"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier1,
            B,
            A,
            *balance_proof_B._asdict().values(),
            balance_proof_update_signature_A,
        ).call({"from": A})

    # Reopen the channel and make sure we cannot use the old balance proof
    channel_identifier2 = create_channel(A, B)[0]
    channel_deposit(channel_identifier2, B, values_B.deposit, A)
    closing_sig = create_close_signature_for_no_balance_proof(B, channel_identifier2)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier2,
            A,
            B,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": B},
    )

    assert channel_identifier1 != channel_identifier2
    with pytest.raises(TransactionFailed, match="TN/update: invalid non-closing sig"):
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier2,
            B,
            A,
            *balance_proof_B._asdict().values(),
            balance_proof_update_signature_A,
        ).call({"from": A})

    # Correct channel_identifier must work
    balance_proof_B2 = create_balance_proof(
        channel_identifier2,
        B,
        values_B.transferred,
        values_B.locked_amounts.locked,
        nonce_B,
        values_B.locksroot,
    )
    balance_proof_update_signature_A2 = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier2,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_B2._asdict(),
    )

    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier2,
            B,
            A,
            *balance_proof_B2._asdict().values(),
            balance_proof_update_signature_A2,
        ),
        {"from": A},
    )


def test_update_channel_event(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    event_handler: Callable,
) -> None:
    """Successful updateNonClosingBalanceProof() emit BALANCE_PROOF_UPDATED events"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit_A = 10
    deposit_B = 10

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)
    channel_deposit(channel_identifier, B, deposit_B, A)
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, 0, 3)
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )
    balance_proof_A = create_balance_proof(channel_identifier, A, 2, 0, 1)
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )

    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )
    txn_hash = call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": B},
    )

    ev_handler.add(
        txn_hash,
        ChannelEvent.BALANCE_PROOF_UPDATED,
        check_transfer_updated(channel_identifier, A, 1, balance_proof_A.balance_hash),
    )
    ev_handler.check()

    # Test event for second balance proof update
    balance_proof_A2 = create_balance_proof(channel_identifier, A, 4, 0, 2)
    balance_proof_update_signature_B2 = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A2._asdict(),
    )
    txn_hash = call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A2._asdict().values(),
            balance_proof_update_signature_B2,
        ),
        {"from": B},
    )

    ev_handler.add(
        txn_hash,
        ChannelEvent.BALANCE_PROOF_UPDATED,
        check_transfer_updated(channel_identifier, A, 2, balance_proof_A2.balance_hash),
    )
    ev_handler.check()



#!/usr/bin/env python
from pathlib import Path

import click
from eth_typing import URI, ChecksumAddress
from eth_utils import encode_hex
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3
from web3.middleware import construct_sign_and_send_raw_middleware

from raiden_contracts.constants import CONTRACT_CUSTOM_TOKEN
from raiden_contracts.contract_manager import ContractManager, contracts_precompiled_path
from raiden_contracts.utils.private_key import get_private_key
from raiden_contracts.utils.signature import private_key_to_address

WEI_TO_ETH = 10 ** 18


@click.command()
@click.option(
    "--rpc-url", help="Ethereum client rpc url", default="http://127.0.0.1:8545", required=True
)
@click.option("--private-key", required=True, type=click.Path(exists=True, dir_okay=False))
@click.option("--token-address", help="address of the token contract", required=True)
@click.option("--amount", help="Amount of tokens to mint", required=True, type=click.INT)
def main(rpc_url: URI, private_key: Path, token_address: ChecksumAddress, amount: int) -> None:
    web3 = Web3(HTTPProvider(rpc_url))
    privkey = get_private_key(private_key)
    assert privkey is not None
    owner = private_key_to_address(privkey)
    web3.middleware_onion.add(construct_sign_and_send_raw_middleware(privkey))
    token_code = web3.eth.get_code(token_address, "latest")
    assert token_code != HexBytes("")
    token_contract = ContractManager(contracts_precompiled_path()).get_contract(
        CONTRACT_CUSTOM_TOKEN
    )
    token_proxy = web3.eth.contract(address=token_address, abi=token_contract["abi"])
    tx_hash = token_proxy.functions.mint(amount).transact({"from": owner})
    print(f"Minting tokens for address {owner}")
    print(f"Transaction hash {encode_hex(tx_hash)}")
    balance = token_proxy.functions.balanceOf(owner).call()
    print(f"Balance of {owner}: {balance}")


if __name__ == "__main__":
    main()



from typing import Callable, Dict

import pytest
from eth_typing import HexAddress
from eth_utils import is_address
from web3 import Web3
from web3.contract import Contract, get_event_data

from raiden_contracts.constants import (
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    EVENT_TOKEN_NETWORK_CREATED,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS
from raiden_contracts.utils.transaction import check_successful_tx


@pytest.fixture()
def get_token_network_registry(deploy_tester_contract: Callable) -> Callable:
    def get(**arguments: Dict) -> Contract:
        return deploy_tester_contract(CONTRACT_TOKEN_NETWORK_REGISTRY, **arguments)

    return get


@pytest.fixture(scope="session")
def token_network_registry_constructor_args(
    web3: Web3, secret_registry_contract: Contract
) -> Dict:
    return {
        "_secret_registry_address": secret_registry_contract.address,
        "_chain_id": web3.eth.chain_id,
        "_settlement_timeout_min": TEST_SETTLE_TIMEOUT_MIN,
        "_settlement_timeout_max": TEST_SETTLE_TIMEOUT_MAX,
        "_max_token_networks": 1,
    }


@pytest.fixture(scope="session")
def token_network_registry_contract(
    deploy_tester_contract: Callable,
    token_network_registry_constructor_args: Dict,
) -> Contract:
    """Deployed TokenNetworkRegistry contract"""
    return deploy_tester_contract(
        CONTRACT_TOKEN_NETWORK_REGISTRY,
        **token_network_registry_constructor_args,
    )


@pytest.fixture(scope="session")
def token_network_registry_contract2(
    deploy_tester_contract: Callable,
    token_network_registry_constructor_args: Dict,
) -> Contract:
    """Another deployed TokenNetworkRegistry contract to which service payment contracts
    should not collaborate."""
    return deploy_tester_contract(
        CONTRACT_TOKEN_NETWORK_REGISTRY,
        **token_network_registry_constructor_args,
    )


@pytest.fixture
def token_network_registry_address(token_network_registry_contract: Contract) -> HexAddress:
    """Address of TokenNetworkRegistry contract"""
    return token_network_registry_contract.address


@pytest.fixture
def add_and_register_token(
    web3: Web3,
    token_network_registry_contract: Contract,
    deploy_token_contract: Callable,
    contracts_manager: ContractManager,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> Callable:
    """Deploy a token and register it in TokenNetworkRegistry"""

    def f(initial_amount: int, decimals: int, token_name: str, token_symbol: str) -> Contract:
        token_contract = deploy_token_contract(initial_amount, decimals, token_name, token_symbol)
        txid = call_and_transact(
            token_network_registry_contract.functions.createERC20TokenNetwork(
                token_contract.address,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ),
            {"from": DEPLOYER_ADDRESS},
        )
        (tx_receipt, _) = check_successful_tx(web3, txid)
        assert len(tx_receipt["logs"]) == 1
        event_abi = contracts_manager.get_event_abi(
            CONTRACT_TOKEN_NETWORK_REGISTRY, EVENT_TOKEN_NETWORK_CREATED
        )
        decoded_event = get_event_data(event_abi, tx_receipt["logs"][0])
        assert decoded_event is not None
        assert is_address(decoded_event["args"]["token_address"])
        assert is_address(decoded_event["args"]["token_network_address"])
        token_network_address = decoded_event["args"]["token_network_address"]
        token_network_abi = contracts_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
        return web3.eth.contract(abi=token_network_abi, address=token_network_address)

    return f



import functools
from collections import defaultdict, namedtuple
from inspect import getframeinfo, stack
from typing import Any, Callable, Dict, List, Optional, cast

from click import echo
from eth_typing.evm import BlockNumber, ChecksumAddress, HexAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3._utils.events import get_event_data
from web3._utils.filters import LogFilter as Web3LogFilter, construct_event_filter_params
from web3._utils.threads import Timeout

# A concrete event added in a transaction.
from web3.types import ABI, ABIEvent, BlockIdentifier

LogRecorded = namedtuple("LogRecorded", "message callback count")
GenesisBlock = BlockNumber(0)


class LogHandler:
    def __init__(self, web3: Web3, address: HexAddress, abi: ABI):
        self.web3 = web3
        self.address = address
        self.abi = abi
        self.event_waiting: Dict[str, Dict[str, LogRecorded]] = {}
        self.event_filters: Dict[str, LogFilter] = {}
        self.event_count: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(lambda: 0))
        self.event_unknown: List[Dict[str, Any]] = []

    def add(
        self,
        txn_hash: str,
        event_name: str,
        callback: Optional[Callable[..., Any]] = None,
        count: int = 1,
    ) -> None:
        caller = getframeinfo(stack()[1][0])
        message = "%s:%d" % (caller.filename, caller.lineno)

        if event_name not in self.event_waiting:
            self.event_waiting[event_name] = {}
            self.event_filters[event_name] = LogFilter(
                web3=self.web3,
                abi=self.abi,
                address=to_checksum_address(self.address),
                event_name=event_name,
                callback=self.handle_log,
            )

        self.event_waiting[event_name][txn_hash] = LogRecorded(
            message=message, callback=callback, count=count
        )

    def check(self, timeout: int = 5) -> None:
        for event in list(self.event_filters.keys()):
            self.event_filters[event].init()

        self.wait(timeout)

    def _handle_waited_log(self, event: Dict[str, Any]) -> None:
        """A subroutine of handle_log
        Increment self.event_count, forget about waiting, and call the callback if any.
        """
        txn_hash = event["transactionHash"]
        event_name = event["event"]
        assert event_name in self.event_waiting
        assert txn_hash in self.event_waiting[event_name]

        self.event_count[event_name][txn_hash] += 1
        event_entry = self.event_waiting[event_name][txn_hash]

        if event_entry.count == self.event_count[event_name][txn_hash]:
            self.event_waiting[event_name].pop(txn_hash)

        # Call callback function with event
        if event_entry.callback:
            event_entry.callback(event)

    def handle_log(self, event: Dict[str, Any]) -> None:
        txn_hash = event["transactionHash"]
        event_name = event["event"]

        if event_name in self.event_waiting:
            if txn_hash in self.event_waiting[event_name]:
                self._handle_waited_log(event)
            else:
                self.event_unknown.append(event)
            if not len(list(self.event_waiting[event_name].keys())):
                self.event_waiting.pop(event_name, None)
                self.event_filters.pop(event_name, None)

    def wait(self, seconds: int) -> None:
        try:
            with Timeout(seconds) as timeout:
                while len(list(self.event_waiting.keys())):
                    timeout.sleep(2)
        except Exception as e:
            echo(e, err=True)
            message = "NO EVENTS WERE TRIGGERED FOR: " + str(self.event_waiting)
            if len(self.event_unknown) > 0:
                message += "\n UNKOWN EVENTS: " + str(self.event_unknown)

            # FIXME Events triggered in an internal transaction
            # don't have the transactionHash we are looking for here
            # so we just check if the number of unknown events we find
            # is the same as the found events
            waiting_events = sum([len(lst) for lst in self.event_waiting.values()])

            if waiting_events == len(self.event_unknown):
                sandwitch_echo(message)
            else:
                raise Exception(
                    message + " waiting_events " + str(waiting_events),
                    " len(self.event_unknown) " + str(len(self.event_unknown)),
                )

    def assert_event(
        self, txn_hash: str, event_name: str, args: List[Any], timeout: int = 5
    ) -> None:
        """Assert that `event_name` is emitted with the `args`

        For use in tests only.
        """

        def assert_args(event: Dict[str, Any]) -> None:
            assert event["args"] == args, f'{event["args"]} == {args}'

        self.add(txn_hash=txn_hash, event_name=event_name, callback=assert_args)
        self.check(timeout=timeout)


def sandwitch_echo(msg: str) -> None:
    echo("----------------------------------", err=True)
    echo(msg, err=True)
    echo("----------------------------------", err=True)


class LogFilter:
    def __init__(
        self,
        web3: Web3,
        abi: ABI,
        address: ChecksumAddress,
        event_name: str,
        from_block: BlockNumber = GenesisBlock,
        to_block: BlockIdentifier = "latest",
        filters: Any = None,
        callback: Optional[Callable[..., Any]] = None,
    ):
        self.web3 = web3
        self.event_name = event_name

        # Callback for every registered log
        self.callback = callback

        event_abi = [i for i in abi if i["type"] == "event" and i["name"] == event_name]
        if len(event_abi) == 0:
            raise ValueError(f"Event of name {event_name} not found")

        self.event_abi = cast(ABIEvent, event_abi[0])
        assert self.event_abi

        filters = filters if filters else {}

        data_filter_set, filter_params = construct_event_filter_params(
            event_abi=self.event_abi,
            abi_codec=web3.codec,
            contract_address=address,
            argument_filters=filters,
            fromBlock=from_block,
            toBlock=to_block,
        )
        log_data_extract_fn = functools.partial(get_event_data, web3.codec, event_abi)

        self.filter: Web3LogFilter = web3.eth.filter(filter_params)
        self.filter.set_data_filters(data_filter_set)  # type: ignore
        self.filter.log_entry_formatter = log_data_extract_fn
        self.filter.filter_params = filter_params

    def init(self, post_callback: Optional[Callable[[], None]] = None) -> None:
        for log in self.get_logs():
            log["event"] = self.event_name
            if self.callback:
                self.callback(log)
        if post_callback:
            post_callback()

    def get_logs(self) -> List[Any]:
        assert self.filter.filter_id is not None
        logs = self.web3.eth.get_filter_logs(self.filter.filter_id)
        formatted_logs = []
        for log in [dict(log) for log in logs]:
            formatted_logs.append(self.set_log_data(log))
        return formatted_logs

    def set_log_data(self, log: Dict[str, Any]) -> Dict[str, Any]:
        log["args"] = get_event_data(
            abi_codec=self.web3.codec, event_abi=self.event_abi, log_entry=log
        )["args"]
        log["event"] = self.event_name
        return log

    def uninstall(self) -> None:
        assert self.web3 is not None
        assert self.filter is not None
        assert self.filter.filter_id is not None
        self.web3.eth.uninstallFilter(self.filter.filter_id)
        del self.filter
